/*
 * Defini un graph dont chaque noeud représente un ensemble de sommets
 * (une communauté) d'un autre graphe
 *
 *  Created on: 21 oct. 2015
 *      Author: dominique
 */

#ifndef SRC_TRIANGLES_GRAPH_HPP_
#define SRC_TRIANGLES_GRAPH_HPP_

#include <functional>

#include "set_vertices_graph.hpp"

using namespace std ;

class triangles_graph : public set_vertices_graph
{

	private :

	template <typename Tvalue>
	static void  compute_triangles(
						    	   graph<Tvalue>& g,
						           Tvertices_map<Tvertices_set>& communauty_vertices,
						           map<Tvertice_identifier, Tvertices_set>& map_vertice_triangles_idents
					              ) ;

	static void compute_triangles_adjacences(
									  	  	 Tvertices_map<Tvertices_set>& communauty_vertices,
											 map<Tvertice_identifier, Tvertice_identifiers_set >& map_vertice_communauty_idents,
											 Tadjacences_map& adjacences
									        ) ;
	public:

		// --------------------- CONSTRUCTEURS ---------------------------------

		triangles_graph() ;

		triangles_graph(
						string aName,
						Tvertices_map<Tvertices_set>& vertices,  // noeud : <id_communaute, {noeud1, noeud2,...}>
						Tadjacences_map& adjacences,
						function<void (Tvertices_set)> display_vertice_value_function,
						function<string (Tvertice_identifiers_set)>  toString_value_function,
						map<Tvertice_identifier, Tvertice_identifiers_set>& map_vertice_communauties_idents
				 	 	) ;

		triangles_graph(
						string aName,
						Tvertices_map<Tvertices_set>& vertices,
						function<void (Tvertices_set)> display_vertice_value_function,
						function<string (Tvertice_identifiers_set)>  toString_value_function
					   ) ;

					// constructeur par copie
		triangles_graph(const triangles_graph& g) ;


		virtual ~triangles_graph() {}


		template <typename Tvalue>
		static triangles_graph*  compute_triangles_graph(graph<Tvalue>& g) ;

} ;


// =========================== IMPLEMENTATIONS =================================


// ----------------------------- CONSTRUCTEURS ---------------------------------

triangles_graph::
triangles_graph() : set_vertices_graph()
{
}

/*
"triangles graph",
								communauty_vertices,
								adjacences,
								graph<Tvertices_set>::display_vertice_identifiers_set,
								graph<Tvertices_set>::toString_vertice_identifiers_set,
							    map_vertice_triangles_idents
							   ) ;
							   */

	 // construit le graphe a partir d'une liste de noeuds et
	 // d'adjacences (avec predecesseurs et successeurs).
	 // ATTENTION : les objets internes sont des copies des paramètres
triangles_graph::
triangles_graph(
				string aName,
				Tvertices_map<Tvertices_set>& vertices,
				Tadjacences_map& adjacences,
				function<void (Tvertices_set)> display_vertice_value_function,
				function<string (Tvertice_identifiers_set)>  toString_value_function,
				map<Tvertice_identifier, Tvertice_identifiers_set>& map_vertice_communauties_idents
				)
				: set_vertices_graph(aName, vertices, adjacences,
						           display_vertice_value_function,
								   toString_value_function,
								   map_vertice_communauties_idents)
{
}

triangles_graph::
triangles_graph(
				string aName,
				Tvertices_map<Tvertices_set>& vertices,
				function<void (Tvertices_set)> display_vertice_value_function,
				function<string (Tvertices_set)>  toString_value_function
				)
				: set_vertices_graph(aName, vertices,
								     display_vertice_value_function,
									 toString_value_function)
{
}

		// constructeur par copie
triangles_graph::
triangles_graph(const triangles_graph& g) : set_vertices_graph(g)
{

}

// créé un graphe des triangles dérivé du graph g
// !!!!! ATTENTION A L'ORIENTATION !!! !!
// !!!!! NE PEUT PAS TROUVER LES TRIANGLES n1, n2, n3 avec !!!!!
// !!! n1-> n2 -> n3 -> n1  (n1 precede n2 precede n3 precede n1) !!!
// ca n'arrive jamais dans les donnees lues par MINERLC
// car les objets sont ordonnées, et les prédecesseurs et successeurs
// sont définis par cet ordre
template <typename Tvalue>
triangles_graph*
triangles_graph::compute_triangles_graph(graph<Tvalue>& g)
{
	Tvertices_map<Tvertices_set> communauty_vertices ;
	map<Tvertice_identifier, Tvertices_set> map_vertice_triangles_idents ;

	triangles_graph::compute_triangles(g,
									   communauty_vertices,
									   map_vertice_triangles_idents) ;

							// creation des adjacences entre communautes
	Tadjacences_map adjacences ;

	triangles_graph::compute_triangles_adjacences(communauty_vertices,
												  map_vertice_triangles_idents,
												  adjacences) ;

	return new triangles_graph(
								"triangles graph",
								communauty_vertices,
								adjacences,
								graph<Tvertice_identifiers_set>::display_vertice_identifiers_set,
								graph<Tvertice_identifiers_set>::toString_vertice_identifiers_set,
							    map_vertice_triangles_idents
							   ) ;
}


/**
 * retourne la liste d'adjacences correspondant aux 3 communautés du graphe courant
 */
void triangles_graph::
compute_triangles_adjacences(
							 Tvertices_map<Tvertices_set>& communauty_vertices,
							 map<Tvertice_identifier, Tvertice_identifiers_set >& map_vertice_communauty_idents,
							 Tadjacences_map& adjacences
							)
{
	Tvertices_map_iterator<Tvertices_set> it 	   = communauty_vertices.begin() ;
	Tvertices_map_iterator<Tvertices_set> it_end = communauty_vertices.end() ;

												// pour chaque noeud triangle
	for ( ; it != it_end ; it++)
	{
			const Tvertice_identifier& triangle_ident = it->first ;
												// it-> first est un numéro de triangle

			Tvertices_set& apexs = it->second.value ;			// ensemble des sommets du triangle

												// sommets du triangle
			const Tvertice_identifier& apex1    = *(apexs.begin()) ;
			const Tvertice_identifier& apex2    = *(++(apexs.begin())) ;
			const Tvertice_identifier& apex3    = *(--apexs.end()) ;

												// on recupere l'ensemble des triangles
												// comprenant ces sommets
			Tvertice_identifiers_set& t_set1  = (map_vertice_communauty_idents)[apex1] ;
			Tvertice_identifiers_set& t_set2  = (map_vertice_communauty_idents)[apex2] ;
			Tvertice_identifiers_set& t_set3  = (map_vertice_communauty_idents)[apex3] ;

												// contriendra les triangles adjacents à triangle
			Tvertice_identifiers_set triangles_adjacents ;


			set_intersection(t_set1.begin(),t_set1.end(), t_set2.begin(),t_set2.end(),
							 inserter(triangles_adjacents, triangles_adjacents.begin()));

			set_intersection(t_set1.begin(),t_set1.end(), t_set3.begin(),t_set3.end(),
							 inserter(triangles_adjacents, triangles_adjacents.begin()));

			set_intersection(t_set2.begin(),t_set2.end(), t_set3.begin(),t_set3.end(),
					         inserter(triangles_adjacents, triangles_adjacents.begin()));

			triangles_adjacents.erase(triangle_ident) ;		// triangle n'est pas adjacent à lui-même

			Tadjacences adja ;
														// répartit les triangles adjacents
														// en prédecesseurs ou successeurs
			Tvertice_identifiers_set_iterator it_t     = triangles_adjacents.begin() ;
			Tvertice_identifiers_set_iterator it_t_end = triangles_adjacents.end() ;

			for (; it_t != it_t_end ; it_t++)
			{
					if (*it_t < triangle_ident)
						adja.predeccessors.insert(*it_t) ;
					else
						adja.successors.insert(*it_t) ;
			}

			adjacences.insert(make_pair(triangle_ident, adja)) ;
	}
}


/**
 * retourne les 3 communauties (ensemble de triangles formés par des noeuds
 * d'un graphe g
 */
template <typename Tvalue>
void  triangles_graph::
compute_triangles(
					graph<Tvalue>& g,
					Tvertices_map<Tvertices_set>& communauty_vertices,
					map<Tvertice_identifier, Tvertices_set>& map_vertice_triangles_idents
				  )		// on parcourt les sucesseurs
{												// de chaque noeud

	Tvertice_identifier c_ident = 0 ;			// identifieur d'un futur noeud d'une communauté

	Tadjacences_map_iterator it      = g.get_adjacences_map().begin() ;
	Tadjacences_map_iterator it_end  = g.get_adjacences_map().end() ;

												// pour chaque noeud du graphe initial
	for (; it != it_end ; it++)
	{
		Tvertice_identifiers_set & successors = (*it).second.successors ;

		if (successors.size() >= 2)
		{										// pour chaque couple de successeurs

			Tvertice_identifiers_set_iterator it2_end = successors.end() ; // fin list des successeurs
			Tvertice_identifiers_set_iterator it1_end = prev(it2_end,1)  ; // avant dernier successeurs

			for (Tvertice_identifiers_set_iterator it1 = successors.begin() ; it1 != it1_end ; it1++)
			{
											// pour chaque successeurs de *it.first
											// on verifie que *it1, *it2
											// sont liés, donc que *it2 fait parti
											// des successeurs de *it1

				for (Tvertice_identifiers_set_iterator it2 = next(it1,1) ; it2 != it2_end ; it2++)
				{
											// on verifie que *it.first, *it1 et *it2 forme un triangle

											// recherche la position de *it1
											// dans la liste d'adjacence

					Tvertice_identifiers_set& successors1 = g.get_successors(*it1) ;

					Tvertice_identifiers_set_iterator it3 = successors1.find(*it2) ;

					if (it3 != successors1.end())
					{
											// *it2 est un successeur de *it1
											// le triangle est formé

						Tvertice<Tvertices_set> vertice ;

						Tvertices_set s = {it->first, *it1, *it2} ;

						vertice.value     = s ;
						vertice.label     = "" ;
						vertice.component = -1 ;

						communauty_vertices.insert(make_pair(c_ident, vertice));

											// on etablie les liens entre les 3 noeuds
											// et le nouveau triangle dont ils sont sommets
						(map_vertice_triangles_idents)[it->first].insert(c_ident) ;
						(map_vertice_triangles_idents)[*it1].insert(c_ident) ;
						(map_vertice_triangles_idents)[*it2].insert(c_ident) ;

						c_ident++ ;
					}
				}
				// fin for it2
			}
			// fin for it1
		}
		// fin if successors.size() > 2
	}
	// fin for it
}
// fin fonction compute_3_communauties


#endif /* SRC_TRIANGLES_GRAPH_HPP_ */
