// minerLC5
/*
** paraminer.cpp
** 
** Made by (Benjamin Negrevergne)
** Email   [firstname].[name]@imag.fr
** 
** Started on  Thu Sep  2 10:21:29 2010 Benjamin Negrevergne
**
** * MINER LOCAL CLOSURES
 *  (by Henry SOLDANO & Guillaume SANTINI (LIPN University Paris 13)
 *  February 2015
 *
 * program modified by Dominique Bouthinon (LIPN, University Paris 13)
 *
 *
 * THIS PROGRAM MUST COMPILED WITH VERSION C++11
 * WITH OPTIONS THAT ALLOWS FOR <functionial>
 *
*/

#include <iostream>
#include <algorithm>
#include <map>
#include <set>
#include <cstdio>
#include <cctype>
#include <cmath>
#include <string>
#include <sstream>

#include "abstract_local_itemsets.hpp"
#include "element.hpp"
#include "pattern.hpp"
#include "database.hpp"
#include "directed_graph.hpp"
#include "display_data.hpp"
#include "types_graph.hpp"
#include "graph.hpp"
#include "miner_local.hpp"
#include "triangles_graph.hpp"
#include "edges_graph.hpp"
#include "utils.hpp"
#include "utils_divers.hpp"
#include "roles.hpp"
#include "statsModularity.hpp"

using std::cout; 
using std::endl;
using std::cerr; 
using std::pair;
				// variables globales

int verbose = 0 ;

unsigned int numero_composante = 0 ;
unsigned int numero_pattern_abstrait = 0 ;
unsigned int nb_spaces = 0 ;
TransactionTable &tt = *new TransactionTable;
Tot &ot = *new Tot;




void printSpaces(string message)
{
	for (int i = 0 ; i < nb_spaces ; i++)
	{
		cerr << " " ;
	}

	cerr << message ;
}

void printSpacesNl(string message)
{
	printSpaces(message) ;
	cerr << endl ;
}


				// fonctions

struct support_sort_cmp_t
{
  const Tot &ot;
  bool operator()(int a, int b)
  {
    return ot[a].size() < ot[b].size(); 
  }
};




// calcule ext_c1, l'extension du pattern c1 (qui inclut le pattern c)
// dans ext_c l'extension de c
void extension(
			   const set_t& c,								// pattern
			   const set_t& c1,								// pattern incluant c
			   const Tot& ot,								// chaque ligne de ot est associée à un item
															// et contient les sommets contenant l'item
			   const Tvertice_identifiers_set& ext_c,		// extension(c)
			   Tvertice_identifiers_set& ext_c1				// extension de c1 dans extension(c)
	  )
{
	set_t attributs_ajoutes ;		// c - c1

	set_difference(
					c1.begin(), c1.end(),
					c.begin(), c.end(),
					inserter(attributs_ajoutes, attributs_ajoutes.begin())
			   	   ) ;

	ext_c1 = ext_c ;

	set_t::iterator it ;

	for (it = attributs_ajoutes.begin() ; it != attributs_ajoutes.end() ; it++)
	{
		// cout << "*it = " << *it << endl ;

		Tvertice_identifiers_set ext ;

		// displaySet("ext_c1", ext_c1) ;
		// displaySet("ot[it]", ot[*it]) ;

		setIntersection(&ext, ext_c1, ot[*it]);
		// displaySet("ext", ext) ;

		ext_c1 = ext ;
	}
}


/**
 * Returns true if parent pattern is the first parent of \pattern
 * according to exclusion list.  Note that given the exclusion list,
 * parent_pattern is not needed to compute this result.
 *
 */
bool parent_pattern_is_first_parent(const set_t &pattern, const set_t &exclusion_list)
{
  assert(is_sorted(pattern));
  assert(is_sorted(exclusion_list));
  /* Check if one element from closed set belong to the exclusion list */
  set_t::const_iterator xlit = exclusion_list.begin();
  const set_t::const_iterator xlend = exclusion_list.end();
  set_t::const_iterator cit = pattern.begin();
  const set_t::const_iterator cend = pattern.end();
  while(xlit != xlend && cit != cend)
  {
    if(*xlit < *cit)
      ++xlit;
    else if (*xlit > *cit)
      ++cit;
    else
    {
      return false;
    }
  }
  return true;
}


		// detecte les triangles de ccc qui doivent être elimines
		// un de leur sommets appartenant à rejected_appexes (sommets éliminés par l'abstraction)
void check_discarded_triangles(
								Tvertice_identifiers_set& rejected_appexes,
								// triangles_graph* gT,
								set_vertices_graph* gT,
								Tvertice_identifiers_set& ccc,
								Tvertice_identifiers_set& discarded_triangles
							  )
{
						// on vérifie pour chaque triangle de ccc
						// qu'aucun de ses sommets n'est dans rejected_appexes

	for (Tvertice_identifiers_set_iterator it_ccc = ccc.begin() ; it_ccc != ccc.end() ; it_ccc++)
	{
											// sommet du triangle courant
		Tvertice_identifiers_set appexes = gT->get_vertice_value(*it_ccc) ;

											// verifie si un sommet du triangle
											// fait partie des sommets rejetes
		bool rejected = false ;

		Tvertice_identifiers_set_iterator it_appex = appexes.begin() ;

		while(it_appex != appexes.end() && !rejected)
		{
			if (find(rejected_appexes.begin(), rejected_appexes.end(), *it_appex)
				!= rejected_appexes.end())
				rejected = true ;
			else
				it_appex++ ;
		}

		if (rejected)
			discarded_triangles.insert(*it_ccc) ;
	}
}

			// met en oeuvre l'abstraction de degré dans la communauté
			// associé au triangle t.
			// nécessite un point fixe
			// (??? visiblement les autres abstractions ne sont pas conservées ???)
void abstract_communauty(
						  Tvertice_identifier t,				// triangle courant
						  graph<string>* g,						// graphe abstrait
						  set_vertices_graph* gT,
						  Tvertice_identifiers_set triangles,	// COPIE (volontaire)
						  int min_degree,
						  int min_communauty_size,
						  Tvertice_identifiers_set& t_rejets,	// triangles rejetés
						  Tvertice_identifiers_set& ccc,		// composante connexe finale
						  	  	  	  	  	  	  	  	  	  	// de t dans gT
						  Tvertice_identifiers_set*& u 			// communaute finale
						 )
{
	graph<string> g1 = *g ;
	// g1.display_graph() ;

	// displaySet("triangles", triangles) ;
	// cout << "t = " << t << endl ;

	// gT->display_graph() ;

	gT->connected_component(t, triangles, ccc) ;

	// displaySet("ccc", ccc) ;


	u = gT->flat_component(ccc) ;

	// displaySet("u", *u) ;

	Tvertice_identifiers_set cc = ccc ;		// sauvegarde 1ere composante connexe

	Tvertice_identifiers_set t_rejects ;

	bool abstraction_ok = true ;
	bool arret = false ;

	while (u->size() >= min_communauty_size && !arret)
	{
		 // g1.display_graph() ;
		 g1.reduce_graph(*u) ; // sous-graphe induit par u

		 // g1.display_graph() ;

		 g1.abstraction_minimal_degree(min_communauty_size, min_degree) ;	// abstraction degre

		 // g1.display_graph() ;

		 Tvertice_identifiers_set v = g1.get_vertice_identifiers_list() ;	// v communaute incluse dans u

		 if (u->size() == v.size())											// point fixe atteint
			 arret = true ;
		 else
		 {
			 if (v.size() >= min_communauty_size)
			 {
				 	 	 	 	 	 	 	 	// v taille valide
				 Tvertice_identifiers_set rejected_appexes ;

				  set_difference(
									u->begin(), u->end(),
									v.begin(), v.end(),
									inserter(rejected_appexes, rejected_appexes.begin())
								 ) ;
				  	  	  	  	  	  	  	  	// on note les triangles de ccc
				  	  	  	  	  	  	  	  	// qui doivent être elimines :
				  	  	  	  	  	  	  	  	// un de leur sommets n'étant plus dans v

				  Tvertice_identifiers_set discarded_triangles ;

				  check_discarded_triangles(rejected_appexes, gT, ccc, discarded_triangles) ;

				  t_rejects.insert(discarded_triangles.begin(), discarded_triangles.end()) ;

				  if (find(discarded_triangles.begin(), discarded_triangles.end(), t)
					  != discarded_triangles.end())
				  {
					  	  	  	  	  	  	  	 // t a été éliminé
					  arret = true ;
					  abstraction_ok = false ;
				  }
				  else
				  {
					  	  	  	  	  	  	  	  // t pas éliminé on continue
					  gT->discard_vertices(discarded_triangles) ;
					  gT->connected_component(t, triangles, ccc) ;
					  u = gT->flat_component(ccc) ;
				  }
			 }
			 else								// v taille insuffisante
			 {
				 arret = true ;
				 abstraction_ok = false ;
			 }
			 // fin if (v.size() >= min_communauty_size)
		 }
		 //  fin if (u->size() == v.size())
	}
	// fin while

	if (u->size() < min_communauty_size || ! abstraction_ok)
	{
		t_rejets = cc ;
		ccc.clear() ;
		u->clear() ;
	}
}

		// affiche les resultats du clos et clos local courants
void affiche_septuples(
						abstraction& a1,
						abstraction& a2,
						Troles_set& roles1,
						Troles_set& roles1_l,
						Troles_set& roles2_l,
						const bool support_demande,
						const set_t& cA,
						const Tvertice_identifiers_set& ext_cA_graphe_initial,
						const Tvertice_identifiers_set& ext_cA,
						const Tvertice_identifiers_set& ext_clos_local_graphe_initial,
						const Tvertice_identifiers_set& ext_clos_local,
						const Tvertice_identifiers_set& component,
						const set_t& clos_local,
						const vector<string>& objects_names,
						const vector<string>& items_names,
						const set_t& permutations
			         )
{
		  // affiche le clos abstrait cA
	outputOriginalPattern2(cA, items_names, permutations) ;
	cout << ", "  ;

	if (support_demande)
	{
		cout << ext_cA_graphe_initial.size() << ", " ;
		cout << ext_cA.size() << ", " ;
		cout << ext_clos_local_graphe_initial.size() << ", " ;
		cout << ext_clos_local.size() << ", " ;
		cout << component.size() << ", " ;
		set<int>::const_iterator it = component.begin() ;
		display_object_name(*it, objects_names) ;
		cout << ", " ;							// *it = index objet
												// ??? prendre un triangle ou arete
												// si graphe dérivé ???
	}
	else
	{
				  // affiche ext(cA) l'extension de cA dans le graphe initial
		display_objects_names(ext_cA_graphe_initial, objects_names) ;
		cout << ", " ;

		  	  	  // affiche ext_a1(c) l'extension abstraite (par a1) de c
	  if (a1.get_role())
		  display_extension_abstraite(roles1, objects_names) ;
	  else
		  display_extension_abstraite(ext_cA, objects_names) ;
	  cout << ", "  ;

			  // affiche ext(l) l'extension de l dans le graphe initial
		display_objects_names(ext_clos_local_graphe_initial, objects_names) ;
		cout << ", " ;

			// affiche ext1(l)
		if (a1.get_role())
			  display_extension_abstraite(roles1_l, objects_names) ;
		else
			  display_extension_abstraite(ext_clos_local, objects_names) ;
		cout << ", "  ;

				  // affiche ext_a2(l) l'extension abstraite (par a1 et a2) de l
		if (a2.get_role())
			  display_extension_abstraite(roles2_l, objects_names) ;
		else
			  display_extension_abstraite(component, objects_names) ;
		cout << ", "  ;
	}
			  // affiche le clos local l
	outputOriginalPattern2(clos_local, items_names, permutations) ;
	cout << endl ;
}



void save_local_graph(graph<string>* initial_graph,graph<string>* gA,
		              bool graphe_oriente, set_t& clos_local,
					  Tvertice_identifiers_set& component,
					  string& data_file, const vector<string>& items_names,
					  const vector<string>& objects_names)
{
	numero_composante++ ;					// definit en global

	graph<string>* gA1 = nullptr ;

	if (graphe_oriente)
	{
		directed_graph<string>* gp =  dynamic_cast<directed_graph<string>* > (initial_graph) ;
		gA1 = new directed_graph<string>(*gp) ;
	}
	else
		gA1 = new graph<string>(*initial_graph) ;

	gA1->reduce_graph(component) ;

	string name3 ;
	get_pattern_name(clos_local, items_names, name3) ;
	// cout << "clos local = " << name3 << endl ;

	Tvertice_identifier object = *(component.begin()) ;

	string object_name =  objects_names.at(object) ;

	stringstream ss;
	ss << numero_pattern_abstrait ;
	string s = ss.str();//return a string with the contents of the stream

	gA1->save_graph_dot(data_file + "_pattern_"+ s +
								  + "_local_" + name3 + + "_cc_"+ object_name , true,true) ;

	delete gA1 ;
}



/**
 *	retourne l'extension globale de pattern à partir de extension
 */
void  extension_globale_pattern(
				 	 	 	const TransactionTable& tt,		// table des transactions
							const Toccurences& extension,	// transactions à considérer,
							set_t& pattern,
							Tvertice_identifiers_set& extension_out,
							const vector<string>& items_names
						  )
{

	// outputOriginalPattern(clos_local, items_names) ;
	// cout << endl ;

	Toccurences::const_iterator it_end = extension.end();

	  	  	  	  	  	  	  	 // pour chaque id transaction de extension
    for(Toccurences::const_iterator it = extension.begin(); it != it_end; it++)
	{
    	// cout << "*it = " << *it << endl ;

		  	  	  	  	  	  	  // tt[*it] est la Transaction designee par it
		 string description ;

		 // get_description_noeud(tt, items_names,	*it, description) ;
		 // cout << "description = " << description << endl ;

		 // cout << "it "

		 // outputOriginalPattern(tt[*it], items_names) ;
		 // cout << endl ;

		 if (inclut(tt[*it], pattern))
		 {
	    						// l est inclus dans la transaction
			  extension_out.insert(*it) ;
		 }
	}
}



// calcule ext(clos_local) l'extension du clos_local et ext_1(clos_local)
// son extension abstraite par a1

void ext_ext1_clos_local(
			bool modularity_on,
		    float modularity,
			Troles_set& roles1_l,
			bool graphe_oriente,
			abstraction& a1,
			graph<string>* g,			  // graphe non abstrait
			graph<string>* initial_graph,
			TransactionTable& tt,
			set_t& clos_local,
			Tvertice_identifiers_set& ext_clos_local,
			Tvertice_identifiers_set& ext_abstraite_clos_local,
			const vector<string>& items_names
		  )
{

	Tvertice_identifiers_set& extension = g->get_vertice_identifiers_list() ;

	// displaySet("extension", extension) ;
	// displaySet("clos_local", clos_local) ;

							// calcule ext_clos_local par rappport
							// à ext(q) = extension, q étant le pattern non
							// abstrait courant avec q inclut dans clos_local
	extension_globale_pattern(tt, extension, clos_local, ext_clos_local, items_names) ;

	// displaySet("ext_clos_local", ext_clos_local) ;

	graph<string>* g1 ;					// copie du graphe non abstrait g

	if (graphe_oriente)
	{
		directed_graph<string>* gp =  dynamic_cast<directed_graph<string>* > (g) ;
		g1 = new directed_graph<string>(*gp) ;
	}
	else
		g1 = new graph<string>(*g) ;

		// sous-graphe induit non abstrait par clos_local
	g1->reduce_graph(ext_clos_local) ;

		// abstraction
	g1->apply_abstractions(a1, -1, initial_graph, roles1_l) ;

	ext_abstraite_clos_local = g1->get_vertice_identifiers_list() ;

	delete g1 ;
}


	//
void calcule_roles_a2(
					  Troles_set& roles2,
					  Tvertice_identifiers_set& component,
		 	 	 	  Troles_set& roles2_l
					 )
{
		Troles_set::const_iterator it_end = roles2.end();

		  	  	  	  	  	  	  	 // pour chaque role
	    for(Troles_set::const_iterator it = roles2.begin(); it != it_end; it++)
		{
	    	setIntersection(&roles2_l[it->first], it->second, component);
		}
}


// calcule les clotures locales sur le graphe initial
int local_closures(
				bool modularity_on,		        	
				float modularity,
				Troles_set& roles1,
				Troles_set& roles2,
				string& data_file,
				bool save_graphs,
				graph<string>*initial_graph,		 // graphe initial
				bool graphe_oriente,
				abstraction& a1,
				abstraction& a2,
	            bool support_demande,
				int depth,
				int min_sup,
				TransactionTable &tt,		// transactions
				graph<string>* g,			// graphe non abstrait du pattern non abstrait courant
				graph<string>* gA,			// graphe courant abstrait par a1 et a2
				const Tvertice_identifiers_set& noeuds_graphe_initial,
				Tot& ot,
				set_t& cA,
				Tvertice_identifiers_set& ext_cA,
				const Tvertice_identifiers_set& ext_cA_graphe_initial,
				SupportTable& support,		// supports dans extension1 des items
										// abstents de closed_pattern
										// (utile uniquement pour connaitre les items
										//  a considérer pour étendre closed_pattern)
				element_t max_element,
				int set_support,
				const vector<string>& objects_names,
				const vector<string>& items_names,
				const set_t& permutations
			  )
{

// cerr << "local closures..." << endl ;
int nb_local_closed_patterns = 0 ;

// gA->visu_graph_dot("graphe_abstrait", true, true) ;

gA->compute_connected_components() ;

// cerr << "end connected components" << endl ;
										// pour chaque composante connexe de g
for (Tconnected_components_iterator it = gA->get_connected_components().begin() ;
  it != gA->get_connected_components().end( ) ; it++)
{

   Tvertice_identifiers_set& component = it->second ;

   int taille = component.size() ;

   if (taille >= min_sup)
   {
	   // displaySet("component", component) ;

										// component est la composante connexe courante
	   nb_local_closed_patterns++ ;

	   SupportTable support_cci(max_element+1, 0);

	   compute_element_support_locale(&support_cci, support, tt, component);

						   // pattern + augmentation présent dans toutes les
						   // transactions de occs_cci.size() ;

	  closure_data_t c_data_cci = {tt, component, support_cci, taille};

	  set_t clos_local  = clo(cA, c_data_cci);		// clos de la localité

						  // calcule ext(l) l'extension de l dans le graphe initial
	  	  	  	  	  	  // et ext_1(l) l'extension abstraite de l dans le graphe initial
	  	  	  	  	  	  // ce à partir de g, le graphe induit par un pattern q incluant l

		  // affiche le clos local l
	  if (save_graphs)
	      save_local_graph(initial_graph, gA, graphe_oriente, clos_local,
	  					  component, data_file, items_names, objects_names) ;


	  Tvertice_identifiers_set ext_clos_local ;
	  Tvertice_identifiers_set ext_abstraite_clos_local ;

	  				  // calcule ext(l) l'extension de l dans le graphe initial
	  			  	  // calcul ext_1(l) extension abstraite de l dans le graphe initial
	  Troles_set roles1_l ;

	  ext_ext1_clos_local(modularity_on, modularity, roles1_l, graphe_oriente, a1, g, initial_graph, tt,
	    				  clos_local, ext_clos_local, ext_abstraite_clos_local,
						  items_names) ;

	  	  	  	  	  // calcule des roles associés à ext2_l = component
	  	  	  	  	  // role_local = component INTER role
	  Troles_set roles2_l ;

	  calcule_roles_a2(roles2, component, roles2_l) ;

	  affiche_septuples(a1, a2, roles1, roles2, roles2_l, support_demande, cA,
			  	  	  	ext_cA_graphe_initial, ext_cA,
						ext_clos_local, ext_abstraite_clos_local,
						component, clos_local,
						objects_names, items_names, permutations) ;
   }
}
// end for each connected component

return nb_local_closed_patterns ;

}
// end local_closures


		// calcule les clotures locales sur le graphe des triangles avec conservation
		// de l'abstraction au sein des communautés
		// ce qui necessite un point fixe
int local_closures_triangles_abstraction(
											bool modularity_on,											
											float modularity,
											Troles_set& roles1,
											Troles_set& roles2,
											graph<string>*initial_graph,		 // graphe initial
											bool graphe_oriente,
											abstraction& a1,
											abstraction& a2,
											bool support_demande,
											int min_degree,			// > 2
											int depth,
											TransactionTable &tt,	// transactions
											graph<string>* g,		// graphe non abstrait du pattern courant
											graph<string>* gA,		// graphe abstrait
											set_vertices_graph* gT,	// graphe des triangles ou des aretes
											Tot& ot,
											set_t& cA,
											Tvertice_identifiers_set& ext_cA,
											const Tvertice_identifiers_set& ext_cA_graphe_initial,
											SupportTable& support,
											element_t max_element,
											int set_support,
											int min_communauty_size,
											const vector<string>& objects_names,
											const vector<string>& items_names,
											const set_t& permutations
										)
{

	// cout << endl << "============================" << endl ;
	// cout << "local_triangles_abstraction" << endl ;
	// g->display_vertices() ;
	// gT->display_graph() ;

	int nb_local_closed_patterns = 0 ;
												// il faut appliquer l'abstraction de degre
												// dans les communautes des composantes connexes

	Tvertice_identifiers_set triangles = gT->get_vertice_identifiers_list() ;	// copie


	while (!triangles.empty())
	{
		// cout << "---" << endl ;
		// displaySet("triangles", triangles) ;
												// pour chaque triangle restant à traiter
		Tvertice_identifiers_set_iterator t = triangles.begin() ;

		// cout << "triangle = " << *t << endl ;

		Tvertice_identifiers_set t_rejets ;		// triangles rejetes
		Tvertice_identifiers_set ccc ;			// composante connexe finale de *t
		Tvertice_identifiers_set* u ;			// communauté finale

		abstract_communauty(*t, gA, gT, triangles, min_degree,
							min_communauty_size, t_rejets, ccc, u) ;

		if (u->size() > 0)
		{										// necessairement |u| >= min_communauty_size
			nb_local_closed_patterns++ ;

			SupportTable support_cci(max_element+1, 0);

			compute_element_support_locale(&support_cci, support, tt, *u);

			int set_support_cci = u->size() ;

			closure_data_t c_data_cci = {tt, *u, support_cci, set_support_cci};
			set_t clos_local  = clo(cA, c_data_cci);


			Tvertice_identifiers_set ext_clos_local ;
			Tvertice_identifiers_set ext_abstraite_clos_local ;

			  // calcule ext(l) l'extension de l dans le graphe initial
			  	 // calcul ext_1(l) extension abstraite de l dans le graphe initial
			Troles_set roles1_l ;

				 // calcule ext(l) l'extension de l dans le graphe initial
			  	  // calcul ext_1(l) extension abstraite de l dans le graphe initial
			ext_ext1_clos_local(modularity_on, modularity, roles1_l, graphe_oriente, a1, g, initial_graph, tt,
							    clos_local, ext_clos_local, ext_abstraite_clos_local,
								items_names) ;

			// calcule des roles associés à ext2_l = component
				  	  	  	  	  // role_local = component INTER role
			Troles_set roles2_l ;

			calcule_roles_a2(roles2, *u, roles2_l) ;

		    affiche_septuples(a1, a2, roles1, roles2, roles2_l,
		    				  support_demande, cA, ext_cA_graphe_initial, ext_cA,
		  		 			  ext_clos_local,
		  					  ext_abstraite_clos_local,	*u,
							  clos_local,
		  					  objects_names,
							  items_names,
							  permutations) ;

				  	  // affiche le clos abstrait cA
		    outputOriginalPattern2(cA, items_names, permutations) ;

		    cout << ", "  ;
		}

		Tvertice_identifiers_set triangles2 ;

										// on retire les triangles elimines
		set_difference (
						 triangles.begin(), triangles.end(),
						 t_rejets.begin(), t_rejets.end(),
						 inserter(triangles2, triangles2.begin())
						) ;

		triangles.clear() ;

										// on retire les triangles déjà traités
		set_difference (
						triangles2.begin(), triangles2.end(),
						ccc.begin(), ccc.end(),
						inserter(triangles, triangles.begin())
						) ;
										// triangles contient les triangles initiaux
										// non traités et non rejetés
		delete u ;
		// delete ccc ;
	}
	// fin while (! triangles.empty())

	return nb_local_closed_patterns ;

}
// end local_closures_triangles_abstraction



	// calcule les clotures locales sur le graphe des triangles
int local_closures_triangles(
								bool modularity_on,								
								float modularity,
								Troles_set& roles1,
								Troles_set& roles2,
								graph<string>*initial_graph,		 // graphe initial
								bool graphe_oriente,
								abstraction& a1,
								abstraction& a2,
								bool support_demande,
								int depth,
								TransactionTable &tt,	// transactions
								graph<string>* g,			// graphe non abstrait du pattern courant
								set_vertices_graph* gT,	// graphe des triangles ou des aretes
								Tot& ot,				// transposition
								set_t& cA,
								Tvertice_identifiers_set& ext_cA,
								const Tvertice_identifiers_set& ext_cA_graphe_initial,
								SupportTable& support,
								element_t max_element,
								int set_support,
								int min_communauty_size,
								const vector<string>& objects_names,
								const vector<string>& items_names,
								const set_t& permutations
							)
{
	int nb_local_closed_patterns = 0 ;

	gT->compute_connected_components() ;
											// pour chaque composante connexe de gT
	for (Tconnected_components_iterator it = gT->get_connected_components().begin() ;
	     it != gT->get_connected_components().end( ) ; it++)
	{
		// cout << "*it = " << *it << endl ;

		Tvertice_identifiers_set component = it->second ;

		// displaySet("component", component) ;

	  	  	  	  	  	  	  	  	  	  	 // recupere l'ensemble des sommets
	  	  	  	  	  	  	  	  	  	  	 // formant les triangles de la composante connexe
	  Tvertice_identifiers_set* communauty = gT->flat_component(component) ;


	  // displaySet("communauty = ", *communauty) ;

	  if (communauty->size() >= min_communauty_size)
	  {
		  	  	  	  	  	  	  	  	  	// la communaute contient assez de sommets
		   nb_local_closed_patterns++ ;

		   SupportTable support_cci(max_element+1, 0);

		   compute_element_support_locale(
										  &support_cci,
										  support,
										  tt,
										  *communauty
										  );

						   // pattern + augmentation présent dans toutes les
						   // transactions de occs_cci ;

		  int set_support_cci = communauty->size() ;

		  closure_data_t c_data_cci = {tt, *communauty, support_cci, set_support_cci};
		  set_t clos_local  = clo(cA, c_data_cci);


		  Tvertice_identifiers_set ext_clos_local ;
		  			Tvertice_identifiers_set ext_abstraite_clos_local ;

		  	  // calcule ext(l) l'extension de l dans le graphe initial
		  	// calcul ext_1(l) extension abstraite de l dans le graphe initial
		  Troles_set roles1_l ;

			  // calcule ext(l) l'extension de l dans le graphe initial
			  // calcul ext_1(l) extension abstraite de l dans le graphe initial
		ext_ext1_clos_local(modularity_on, modularity, roles1_l, graphe_oriente, a1, g, initial_graph, tt,
							clos_local, ext_clos_local, ext_abstraite_clos_local,
							items_names) ;


		// calcule des roles associés à ext2_l = component
						  	  	  	  	  // role_local = component INTER role
		Troles_set roles2_l ;

		calcule_roles_a2(roles2, *communauty, roles2_l) ;

	    affiche_septuples(a1, a2, roles1, roles2, roles2_l,
						support_demande, cA, ext_cA_graphe_initial, ext_cA,
							ext_clos_local,
							ext_abstraite_clos_local, *communauty, clos_local,
							objects_names, items_names, permutations) ;
	  }

	  delete communauty ;
	}
	// end for each connected component

	return nb_local_closed_patterns ;
}
// end local_closures_triangles

						// Algo inspire de PARAMINER (enum_clos)
size_t expand(
				Troles_set& roles,
				string data_file,
				int m,						// nombre arretes graph initial
				bool save_graphs,
				statsModularity& statsMod,
				int& nec,					// nombre de necessaires
				map<Tvertice_identifier, Tvertice_identifiers_set>& c_edges,
				bool output_pattern,		// true si le pattern courant est ok
				bool modularity_on,
				float modularity,
				const bool support_demande,
				abstraction& a1,
				abstraction& a2,
				const bool a2_active,
				const bool communauty_abstraction,
				TransactionTable& tt,				 // transactions
				Tot& ot,
				const Tvertice_identifiers_set& noeuds_graphe_initial,
				const Tvertice_identifiers_set& extension_cA_precedent_graphe_initial,
													 // noeuds du GRAPHE INITIAL satisfaisant le pattern
													 // clos abstrait précédent
													 // (initialement ensemble des noeuds du courant).

				graph<string>*initial_graph,		 // graphe initial
				graph<string>* g,					 // graphe non abstrait
				graph<string>* gA,					 // graphe abstrait courant 
				                                     // satisfaisant cA_precedent + augmentation
													 // (initialement pattern vide)
				bool graphe_oriente,				 // true => graphe oriente
				const set_t& cA_precedent,    	   	 // clos abstrait precedent (pattern vide pour premier appel)
				const element_t augmentation,     	 // augmentation envisage du clos abstrait precedent
													 // (-1 si depth == 0)
				int depth,
				const set_t &parent_el,         	 // vide pour depth = 0
				const set_t &el_tail,           	 // vide pour depth = 0
				const int min_sup,
				const bool local,
				bool build_triangles_graph,
				bool build_edges_graph,
				int  k_clique,
				int  k_crown,
				const vector<string>& objects_names,
				const vector<string>& items_names,
				const set_t& permutations,
				int& nb_closed_patterns,
				int& nb_rejected_closed_patterns
			)
{

 if (verbose > 0)
	 nb_spaces += 3 ;

 set_t pattern(cA_precedent);
 set_t exclusion_list(parent_el.size()+el_tail.size());

 copy(parent_el.begin(), parent_el.end(), exclusion_list.begin());
 copy(el_tail.begin(), el_tail.end(), exclusion_list.begin()+parent_el.size());
 assert(is_sorted(parent_el));

 sort(exclusion_list.begin(), exclusion_list.end());

 // displayVector("parent_el", parent_el) ;
 // displayVector("el_tail", el_tail) ;
 // displayVector("exclusion_list", exclusion_list) ;

									  // extension abstraite courante correspondant aux
 	 	 	 	 	 	 	 	 	  // sommets contenant cA_precedent + augmentation
 	 	 	 	 	 	 	 	 	  // ATTENTION POINTEUR VERS LES NOEUDS (PAS UNE COPIE)

									// extension abstraite
  Tvertice_identifiers_set& extensionA = gA->get_vertice_identifiers_list() ;

  // displaySet("extensionA", extensionA) ;

  element_t max_element = tt.max_element;

  SupportTable support(max_element+1, 0);

  	  	  	  	  	  	  	  	  	  // calcule les supports des items qui
  	  	  	  	  	  	  	  	  	  // apparaissent dans les noeuds de gA
  	  	  	  	  	  	  	  	  	  // (les labels des noeuds de gA satisfont "pattern + augmentation")
  	  	  	  	  	  	  	  	  	  // (pour depth = 0 calcule les supports
  	  	  	  	  	  	  	  	  	  // des items apparaissant dans les noeuds de gA satisfaisant
  	  	  	  	  	  	  	  	  	  // le "pattern vide"
  compute_element_support(&support, tt, extensionA);

  if (depth > 0)
  {
  	  	  	  	  	  	  	  	  	  // ajout de l'augmentation au pattern courant
	  pattern.push_back(augmentation);
  }

  	  	  	  	  	  	  	  	  	  // pattern n'est plus ni clos (ni abstrait)
  int set_support ;

  if (depth > 0)
  {
  	  	  	  	  	  	  	  	  	  // initialise le support du pattern
  	  	  	  	  	  	  	  	  	  // parent + augmentation
  	  	  	  	  	  	  	  	  	  // avec le support de l'augmentation
	  set_support = support[augmentation];

									  // remet les supports des items appartenant
									  // à pattern + augmentation à 0
	  	  	  	  	  	  	  	  	  // (ca evitera a la cloture de les considerer
	  	  	  	  	  	  	  	  	  //  comme items à ajouter)
	  for(set_t::const_iterator it = pattern.begin(); it != pattern.end(); ++it)
	  {
		if(*it <= max_element)
		  support[*it] = 0;
	  }
  }
  else
  {
	  set_support = gA->nb_vertices() ;
  }
  	  	  	  	  	  	  	  	  	  // support[item] non null =>
  	  	  	  	  	  	  	  	  	  // (item n'est pas dans pattern)
  	  	  	  	  	  	  	  	  	  // pattern_augmentation + item est present
  	  	  	  	  	  	  	  	  	  // dans les transactions de occs .

// cout << endl ;
// cout << " ============================================" << endl ;

						  /***************************************/
						  /* Compute pattern closure             */
  	  	  	  	  	  	  /* patterns dont le support est egal   */
  	  	  	  	  	  	  /* a support[pattern_augmentation]     */
						  /***************************************/

  if (verbose > 0)
  {
  	 string nameP ;
  	 get_pattern_name(pattern, items_names, nameP) ;
  	 if (nameP.empty())
  		 printSpacesNl("pattern considere : pattern vide") ;
  	 else
  		printSpacesNl("pattern considere : " + nameP) ;
   }


  closure_data_t c_data = {tt, extensionA, support, set_support};
  set_t cA = clo(pattern, c_data);			// cA = pattern clos abstrait de gA
  sort(cA.begin(), cA.end());				// dont l'extension abstraite est extensionA

  	  	  	  	  	  	  	  	  	  	    // gA graphe abstrait dont les (labels)
  	  	  	  	  	  	  	  	  	  	  	// des noeuds conntiennent le clos abstrait cA

  if (verbose > 0)
  {
 	 string name3 ;
 	 get_pattern_name(cA, items_names, name3) ;
 	 if (name3.empty())
 		 printSpacesNl("pattern clos considere : pattern vide") ;
 	 else
 		printSpacesNl("pattern clos considere : " + name3) ;
  }


						  /******************************************************************/
						  /* Check if parent pattern is the first parent of closed_pattern. */
						  /******************************************************************/
  if(!parent_pattern_is_first_parent(cA, exclusion_list))
  {
	if (modularity_on)
	{
		statsMod.decNbLme() ;

		if (output_pattern)
		{
			statsMod.decNbLm() ;
		}

		// cout << "lme = " << statsMod.getNbLme() << endl ;
		// cout << "lm = "  << statsMod.getNbLme() << endl ;
	}

	if (verbose > 0)
	{
		printSpacesNl("pas premier parent ") ;
		nb_spaces -= 3 ;
	}

    return 0;
  }



int nb_local_closed_patterns2 = 0 ;

			// calcule l'extension (non abstraite) du clos abstrait cA courant
			// dans le graphe initial
Tvertice_identifiers_set extension_cA_graphe_initial ;
extension(cA_precedent, cA, ot, extension_cA_precedent_graphe_initial, extension_cA_graphe_initial) ;

if (output_pattern)
{

	if (verbose > 0)
	{
		printSpacesNl("clos ok ") ;
	}

				// le pattern courant doit etre sorti on le traite
	nb_closed_patterns++ ;
	numero_pattern_abstrait++ ;

	if (save_graphs)
	{
		string name3 ;
		get_pattern_name(cA, items_names, name3) ;

		stringstream ss;
		ss << numero_pattern_abstrait ;
		string s = ss.str();//return a string with the contents of the stream

		gA->save_graph_dot(data_file + "_pattern_" +  s + "_" + name3, true, true) ;
	}

  // gA->visu_graph_dot("graphe_abstrait", true, true) ;

  	Tvertice_identifiers_set adj ;
  	// gA->adjacent_vertices(2, adj) ;
  	// displaySet("adj_2", adj) ;



  	  	  	  	  	  	  	  	  // le fait de decomposer le graphe
  	  	  	  	  	  	  	  	  // des triangles en composantes connexes
  	  	  	  	  	  	  	  	  // et d'eliminer certaines d'entre elles
  	  	  	  	  	  	  	  	  // ne doit pas eliminer de noeuds dans
  	  	  	  	  	  	  	  	  // le graphe initial g

   	  	  	  	  	  	  	  	  // clotures locales
  if (local)
  {
	  	  	  	  	// on met les supports des items de cA a zero
	  	  	  	  	// (afin de ne pas les considerer comme items
	  	  	  	  	//  a ajouter dans les clos locaux)

	  for(set_t::const_iterator it = cA.begin(); it != cA.end(); ++it)
	  {
			if(*it <= max_element)
				support[*it] = 0;
	  }

	  graph<string>* gA2   = NULL ;

	  Troles_set roles2 ;

	  if (a2_active)
	  {
		 gA2  = new graph<string>(*gA);		// graphe associe a l'abtraction a2
	  	 	 	 	 	 	 	 	 	 	 // on applique les abstractions de a2
	     gA2->apply_abstractions(a2, min_sup, initial_graph, roles2) ;

	     // gA2->visu_graph_dot("gA2", true, true) ;
	  }
	  else
	  {
		  gA2 = gA ;							// on conserve le graphe gA
		  roles2 = roles ;
	  }

	  bool output_pattern3 ;
	  map<Tvertice_identifier, Tvertice_identifiers_set> c_edges3 ;
	  statsModularity statsMod3 ;

	  if (gA2->graph_ok(min_sup, statsMod3, modularity_on, modularity, initial_graph, initial_graph,
			            c_edges3, output_pattern3))
	  {
		  // gA2->visu_graph_dot("gA2", true) ;

		  if (build_triangles_graph || k_clique > 0 || k_crown > 0)	// on passe au graphe des triangles ou des aretes
		  {
			  set_vertices_graph* gT2 = nullptr ;

			  if (build_edges_graph || k_clique > 0 || k_crown > 0)
			  {
				  	  	  	  	  	  	  	  	  	  	  	  // graphe des aretes
				  gT2 = edges_graph::compute_edges_graph(*gA2) ;

				  // gT2->visu_graph_dot("gT2", true, true) ;

				  edges_graph* gT3 = (edges_graph*) gT2 ;	  // cast !

				  	  	  	  	  	  	  	  	  	  	  	  // gT3 pointe sur gT2 !
				  if  (k_clique > 0 || k_crown >0)
					  gT3->keep_k_cliques_ou_k_couronnes(k_clique, k_crown, *gA2) ;
						// cast !

				  // gT3->visu_graph_dot("gT3", true, true) ;

				  // gT3->abstraction_3_clique(*gA2) ;
				  	  	  	  	  // desormais les communautes de gT2 sont
				  	  	  	  	  // equivalentes a celles du graphe des triangles
				  	  	  	  	  // derive de gA2
			  }
			  else
			  {
				  gT2 = triangles_graph::compute_triangles_graph(*gA2) ;
			      // gt->visu_graph_dot("triangles_graph", true) ;
			  }

			  int min_communauty_size ;			// taille minimale de la communaute

			  if (a2.get_connected_components() > min_sup)
			  		min_communauty_size = a2.get_connected_components() ;
			  	else
			  		min_communauty_size = min_sup ;

			  int min_degree ;

			  if (a2.get_degree() > 2)
			  		min_degree = a2.get_degree() ;
			  else
			  		min_degree = 2 ;

			  if (min_degree > 2 && communauty_abstraction)

				  	  	  	  	  	  	  // on verifie que les abstractions
				  				  	  	  // sont preserves dans les commmunautes associees
				  				  	  	  // aux composantes connexes du graphe des triangles
				  nb_local_closed_patterns2 = local_closures_triangles_abstraction(
						  	  	  	  	  	  	  	  	  	  	  	  	  	  	    modularity_on,
																				    modularity,
						  	  	  	  	  	  	  	  	  	  	  	  	  	  	    roles,
						 					  	  	  	  	  	  	  	  	  	    roles2,
						  	  	  	  	  	  	  	  	  	  	  	  	  	  	    initial_graph,
						  	  	  	  	  	  	  	  	  	  	  	  	  	  	  	graphe_oriente,
																					a1,
																					a2,
						                                                            support_demande,
						  	  	  	  	  	  	  	  	  	  	  	  	  	  	  	min_degree,
																					depth,
																					tt,
																					g,
																					gA2,
																					gT2,
																					ot,
																					cA,
																					extensionA,
																					extension_cA_graphe_initial,
																					support,
																					max_element,// le noeud courant respecte la contrainte de modularite,
																					set_support,
																					min_communauty_size,
																					objects_names,
																					items_names,
																					permutations
																				  ) ;
			  else						// on ne verifie pas la préservation des abstractions
				  	  	  	  	  	  	// dans les commmunautes associees
				  	  	  	  	  	  	// aux composantes connexes du graphe des triangles
				  nb_local_closed_patterns2 = local_closures_triangles(
																	    modularity_on,
						  	  	  	  	  	  	  	  	  	  	  	    modularity,
						  	  	  	  	  	  	  	  	  	  	  	    roles,
						 					  	  	  	  	  	  	  	roles2,
						  	  	  	  	  	  	  	  	  	  	  	  	initial_graph,
						  	  	  	  	  	  	  	  	  	  	  	    graphe_oriente,
																		a1,
																		a2,
						                                                support_demande,
																		depth,
																		tt,
																		g,
																		gT2,
																		ot,
																		cA,
																		extensionA,
																		extension_cA_graphe_initial,
																		support,
																		max_element,
																		set_support,
																		min_communauty_size,
																		objects_names,
																		items_names,
																		permutations
																	  ) ;

			  delete gT2 ;
		  }
		  else								// pas de passage au graphe des triangles			  	  	  	  	  	  	 
		  {
			  nb_local_closed_patterns2  = local_closures(
					  	  	  	  	  	  	  	  	  	 modularity_on, 
														 modularity,
					  	  	  	  	  	  	  	  	  	  roles,
					  	  	  	  	  	  	  	  	  	  roles2,
					  	  	  	  	  	  	  	  	  	  data_file,
														  save_graphs,
					  	  	  	  	  	  	  	  	  	  initial_graph,
					  	  	  	  	  	  	  	  	  	  graphe_oriente,
														  a1,
														  a2,
					                                      support_demande,
					                                      depth,
					  	  	  	  	  	  	  	  	  	  min_sup,
														  tt,
														  g,
														  gA2,
														  noeuds_graphe_initial,
														  ot,
														  cA,
														  extensionA,
														  extension_cA_graphe_initial,
														  support,
														  max_element,
														  set_support,
														  objects_names,
														  items_names,
														  permutations
														 ) ;
		  }

		  if (nb_local_closed_patterns2 == 0)
		  {
			  if (a2_active)
			  {
			 			  delete gA2 ;
			  }
			  nb_rejected_closed_patterns++ ;

			  // nb_spaces -= 3 ;

			  return 0 ;
		  }
	  }
	  // fin if (g->nb_vertices) >= min_sup

	  if (a2_active)
	  {
	 	  delete gA2 ;
	  }
  }
  else
  {											// mode global
	  	  	  	  // affiche le clos abstrait
	  outputOriginalPattern(cA, items_names) ;
	  cout << ", "  ;

	  if (support_demande)
	  {
		  cout << extension_cA_graphe_initial.size() << ", " ;
		  cout << gA->nb_vertices() << endl ;
	  }
	  else
	  {
					  // affiche l'extension (non abstraite) du clos abstrait cA courant
					  // dans le graphe initial
		  display_objects_names(extension_cA_graphe_initial, objects_names) ;
		  cout << ", " ;

		  	  	  	  // affiche l'extension abstraite (l'ensemble des objets du graphe abstrait)
		  if (a1.get_role())
			  display_extension_abstraite(roles, objects_names) ;
		  else
			  display_extension_abstraite(gA->get_vertice_identifiers_list(), objects_names) ;

		  cout << endl  ;
	  }

	  if (save_graphs)
	  {
		  string name ;
		  get_pattern_name(cA, items_names, name) ;
		  gA->save_graph_dot(data_file + "_abstrait_" + name, true, true) ;
	  }
  }
  // fin else (mode global)

}
// fin de if output_pattern (fin traitement du clos sorti)
else
{
	if (verbose > 0)
		printSpacesNl("clos satisfait pas lm (pas traite), poursuite specialisations") ;
}


	  	  //TODO could be improved,
for(set_t::const_iterator it = cA.begin(); it != cA.end(); ++it)
{
	if(*it <= tt.max_element)
	support[*it] = 0;    //By doing this we remove from db the elements
						// that already belong to the set.
}

						  /*************************************************/
						  /* Computes current closed pattern augmentations */
						  /*************************************************/
  set_t augmentations;

  /* to store the value returned by membership foreach element. (Will
     be passed to the rec. call.) */
  set_t augmentations_membership_retval(max_element+1); 

  for(int item = 0; item <= max_element; item++)
  {
    if(support[item] != 0)
    {
      if(!set_member(exclusion_list, item))
    	  augmentations.push_back(item) ;
    }
    // fin if (support[item]...)
  }
  // fin for chaque item


  int nec1 = 0 ; 			// nombre de necessaires parmi les noeuds fils

  if(augmentations.size() > 0)
  {
    set_t new_exclusion_list;
    set_t new_el_tail;
    TransactionTable *new_tt;
    Tot *new_ot;

				  /**************************/
				  /* Reuses parent dataset. */
				  /**************************/
    new_tt = &tt;
    new_ot = &ot;
      	  	  	  /* Re-compute the new tids (this can be improved) */
    new_exclusion_list = parent_el;
    new_el_tail = el_tail;

    support_sort_cmp_t support_sort = {*new_ot};      
    std::sort(augmentations.begin(), augmentations.end(), support_sort); 

							  /*****************************/
							  /* Recursive call to expand. */
							  /*****************************/

    						  // pour chaque augmentation possible
      set_t::const_iterator e_end = augmentations.end();

      // int i = 0 ;
      // cout << "nb augmentations " << augmentations.size() << endl ;

      for(set_t::const_iterator e = augmentations.begin(); e != e_end; ++e)
      {

    	  set_t augment ;

    	  augment.push_back(*e) ;

    	  if (verbose > 0)
    	  {
    		  string nameP ;
    		  get_pattern_name(augment, items_names, nameP) ;
    		  printSpacesNl("augmentation = " + nameP) ;
    	  }


    	   	  	  	  	  	  	  	  	  	  // cree un nouveau graphe abstrait
    	  graph<string> * gA1 ;

    	  if (graphe_oriente)
    	  {
    	  		directed_graph<string>* gp =  dynamic_cast<directed_graph<string>* > (gA) ;
    	  		gA1 = new directed_graph<string>(*gp) ;
    	  }
    	  else
    		  	gA1 = new graph<string>(*gA);


    	  Tvertice_identifiers_set extensionA1;
    	  // displaySet("extensionA", extensionA) ;

    	  	  	  	  	  	  // on reduit l'extension courante aux objets (noeuds, transactions)
    	  	  	  	  	  	  // de extensionA qui possedent l'augmentation *e_it
		  setIntersection(&extensionA1, extensionA, (*new_ot)[*e]);

		  	  	  	  	  	  // new_occs contient les identifiants des transactions
		  	  	  	  	  	  // correspondant à l'extension abstraite courante
		  	  	  	  	  	  // dont le pattern COMPREND AUSSI l'augmentation  courante *c_it

		  // displaySet("extensionA1", extensionA1) ;
		  // cout << endl ;

		  	  	  	  	  	  // sous-graphe induit non abstrait par nouveau pattern courant
		  	  	  	  	  	  // cA + augmentation
		  gA1->reduce_graph(extensionA1) ;

		  Troles_set roles1 ;
		  	  	  	  	  	  // abstraction
		  gA1->apply_abstractions(a1, min_sup, initial_graph, roles1) ;

		  bool output_pattern1 ;
		  map<Tvertice_identifier, Tvertice_identifiers_set> c_edges1 = c_edges ;

		  if (gA1->graph_ok(min_sup, statsMod, modularity_on, modularity, initial_graph,
				            gA, c_edges1, output_pattern1))
		  {

			  graph<string>* g1 = nullptr ;

			  if (local)
			  {				  	  	  	  	  	  	  	  	  // copie du graph g
			  		if (graphe_oriente)
			   		{
			   			directed_graph<string>* gp =  dynamic_cast<directed_graph<string>* > (initial_graph) ;
			   			g1 = new directed_graph<string>(*gp) ;
			   		}
			   		else
			   			g1 = new graph<string>(*initial_graph) ;

					Tvertice_identifiers_set& extension1 = g1->get_vertice_identifiers_list() ;
			    	Tvertice_identifiers_set extension_g1;

			    	set_t cA_e = cA ;
			    	cA_e.push_back(*e) ;

			    	extension_globale_pattern(tt,	extension1,	cA_e, extension_g1, items_names) ;

					  	  	  	  	  	  // sous-graphe induit non abstrait par nouveau pattern courant
					  	  	  	  	  	  // cA + augmentation
					g1->reduce_graph(extension_g1) ;
			  }

			 int nec2 = 0;

			      // nbTabs += 5 ;
			 nb_local_closed_patterns2 += expand(
					 	 	 	 	 	 	 	 	  roles1,
						  	  	  	  	  	  	  	  data_file,
													  m,
						  	  	  	  	  	  	  	  save_graphs,
													  statsMod,
													  nec2,
													  c_edges1,
													  output_pattern1,
													  modularity_on,
													  modularity,
						                              support_demande,
						                              a1,
													  a2,
													  a2_active,
													  communauty_abstraction,
													  *new_tt,
													  *new_ot,
													  noeuds_graphe_initial,
													  extension_cA_graphe_initial,
													  initial_graph,
													  g1,			    // graphe non abstrait courant
													  gA1,				// graphe abstrait courant
													  graphe_oriente,
													  cA,				// pattern clos courant du graphe abstrait precedent
													  *e,				// augmentation de cA
													  depth+1,
													  new_exclusion_list,
													  new_el_tail,
													  min_sup,
													  local,
													  build_triangles_graph,
													  build_edges_graph,
													  k_clique,
													  k_crown,
													  objects_names,
													  items_names,
													  permutations,
													  nb_closed_patterns,
													  nb_rejected_closed_patterns);

								  /* insert the current augmentation into the exclusion list for the next calls.*/
				  new_el_tail.push_back(*e);

				  if (local)
					  delete g1 ;

				 if (modularity_on)
					 nec1 = nec1 + nec2 ;
		  }
		  // fin if graph_ok
		  else
		  {
			  if (verbose > 0)
				  printSpacesNl("augmentation improductive") ;
		  }

		  delete gA1 ;
		}
        // fin for
   }
  // fin if augmentation.size() > 0


  if (verbose > 0)
  {
	  	 string name3 ;
	  	 get_pattern_name(cA, items_names, name3) ;

	  	 if (name3.empty())
	  		 printSpacesNl("fin traitement du pattern clos : pattern vide") ;
	  	 else
	  		printSpacesNl("fin traitement du pattern clos : " + name3) ;
  }

  if (modularity_on)
  {
	  if(
        	  nec1 > 0				// la descendance du noeud courant contient au moins
     		   	   	   	   	        // un pattern respectant la contrainte de modularite et min_sup
     			   	   	   ||
 			  output_pattern		// le noeud courant respecte la contrainte de modularite et min_sup
 		)
     		  nec = 1 + nec1 ;
      else
     		  nec = nec1 ;

	  if (nec1 == 0 && output_pattern)	// descendance ne contient pas un noeud ok
	  {									// noeud courant ok => noeud frontiere
		  statsMod.incF() ;
	  }
  }

  if (verbose > 0)
	  nb_spaces -= 3 ;

  return nb_local_closed_patterns2 ;
}


/* ========================================================================== */

				// calcul les patterns clos abstraits
int abstract_local_itemsets(
		                    string data_file,
							bool save_graphs,
							statsModularity& statsMod,
						    bool modularity_on,
							float modularity,
							bool support_demande,
							abstraction& a1,
							abstraction& a2,
							bool a2_active,
							bool communauty_abstraction,
							int min_sup,
							bool local,
							bool build_triangles_graph,
							bool build_edges_graph,
							int  k_clique,
							int  k_crown,
							graph<string>* initial_graph,
							bool graphe_oriente,
							const vector<string>& objects_names,
							const vector<string>& items_names,
							const set_t& permutations,
							int& nb_closed_patterns,
							int& nb_rejected_closed_patterns
						    )
{
									// abstraction intermédaire :
				 					// 1) d'abord A1 puis cloture c sur A1
									// 2) puis (A1, A2) sur les localités

	// cerr << endl << "BEGIN PARAMINER WITH ABSTRACTION(S) AND LOCAL CLOSURES.." << endl ;

	cout << "# ===================================================================" << endl ;

	// int m = initial_graph->nb_edges() ;		// # aretes graphe initial

	int nb_local_closed_patterns = 0 ;

	if (tt.size() < min_sup)				// inutile de continuer si pas assez de transactions
	{
		    // cerr << endl <<  "END PARAMINER LOCAL CLOSURES." << endl << endl ;
			return nb_local_closed_patterns ;
	}

	set_t empty_set;
	set_t exclusion_list;
	set_t exclusion_list_tail;

										// copie noeuds du graphe initial
	Tvertice_identifiers_set noeuds_graphe_initial = initial_graph->get_vertice_identifiers_list() ;

	Occurence all_tids(tt.size());

	for(unsigned int i = 0; i < tt.size(); i++)
		all_tids[i]=i;

	if (save_graphs)
	{
		initial_graph->save_graph_dot(data_file + "_graphe_initial", true, true) ;
	}

	graph<string>* gA ;
											// copie du graphe initial
	if (graphe_oriente)
	{
		directed_graph<string>* gp =  dynamic_cast<directed_graph<string>* > (initial_graph) ;
		gA = new directed_graph<string>(*gp) ;
	}
	else
		gA = new graph<string>(*initial_graph) ;

	Troles_set roles ;


	// gA->visu_graph_dot("graphe_initial", true, true) ;

	 	 	 	 	 // abstraction
	gA->apply_abstractions(a1, min_sup, initial_graph, roles) ;

    bool output_pattern ;
    map<Tvertice_identifier, Tvertice_identifiers_set> c_edges;

    // cout << "pattern envisage " << endl ;
    // cout << "{}" << endl ;

	if (!gA->graph_ok(min_sup, statsMod, modularity_on, modularity, initial_graph,
					  initial_graph, c_edges, output_pattern))
	{
	    // cerr << endl <<  "END PARAMINER LOCAL CLOSURES." << endl << endl ;
	    delete gA  ;
	    delete initial_graph   ;
	    return nb_local_closed_patterns ;
	}

	graph<string>* g = nullptr ;

	if (local)
	{
												// copie du graphe initial
												// (utilisee hors abstraction)
		if (graphe_oriente)
		{
			directed_graph<string>* gp =  dynamic_cast<directed_graph<string>* > (initial_graph) ;
			g = new directed_graph<string>(*gp) ;
		}
		else
			g = new graph<string>(*initial_graph) ;
	}

	int m = 0 ;
	int nec = 0 ;				// nombre de necessaires

	nb_local_closed_patterns  = expand(
										roles,
										data_file,
										m,
			                            save_graphs,
										statsMod,
										nec,
										c_edges,
			                            output_pattern,
										modularity_on,
										modularity,
										support_demande,
										a1,
										a2,
										a2_active,
										communauty_abstraction,
										tt,
										ot,
										noeuds_graphe_initial,
										noeuds_graphe_initial,
										initial_graph, 
										g,
										gA,
										graphe_oriente,
										empty_set,
										-1,  // pattern_augmentation (rien au depart)
										0,   // depth
										exclusion_list,
										exclusion_list_tail,
										min_sup,
										local,
										build_triangles_graph,
										build_edges_graph,
										k_clique,
										k_crown,
										objects_names,
										items_names,
										permutations,
										nb_closed_patterns,
										nb_rejected_closed_patterns
									  ) ;

	if (modularity_on)
    {
		statsMod.setNec(nec) ;
    }

	delete gA ;

	if (local)
		delete g  ;

    return nb_local_closed_patterns;
}

