# -*- coding: utf8 -*-

import sys
import os
import re

def indexDict( d_list) :
	dico = dict([ (x,i) for i,x in enumerate( d_list) ])
	return dico

def run( args) :

	# -------------------------------------
	# Read input data from .ri file

	fin = open( args[1],'r')
	ri_data = fin.readlines()
	fin.close()
	
	# -------------------------------------
	# Read the list of objects

	objIndex=indexDict( map(lambda s: s.strip(), ri_data[4].split('|')))
	nb_obj = len( objIndex) 

	# -------------------------------------
	# Read the itemsets

	itemset = {}
	for i in xrange( nb_obj) :
		itemset[i] = [ j for j,x in enumerate(re.findall(r"([01])\s*",ri_data[6+i].strip())) if '1' == x ]

	# -------------------------------------
	# Read adjacencies

	adjacencies={}
	for i in xrange( len( ri_data) - 8 - nb_obj - 1) :
		m = re.match( r"^([\w\-@]+)\s*(.*)", ri_data[ i + 8 + nb_obj])
		ref = m.group(1)
		adjacencies[objIndex[ref]] = []
		for o in re.findall(r"([\w\-@]+)\s*,?",m.group(2)) :
			adjacencies[objIndex[ref]].append(objIndex[o])
		adjacencies[objIndex[ref]].sort()


	# -------------------------------------
	# Write outfile Header

	fout = open( args[2],'w')

	fout.write("# " + ri_data[1].strip() + " / " + ri_data[3].strip() +"\n")	# write header
	fout.write(ri_data[4])		# write list of objects					
	fout.write(ri_data[5])		# write list of items

	# -------------------------------------
	# Write sparse list for each object itemset

	for i in xrange( nb_obj) :
		fout.write( str(i))
		if 0 < len( itemset[i] ) :
			fout.write( ' '+','.join(map(lambda i: str(i),itemset[i])))
		fout.write( '\n')
	
	fout.write("#\n")

	# -------------------------------------
	# Write sparse list for each object adjacencies
	for i in xrange( nb_obj) :
		fout.write( str(i))
		try:
			if 0 < len( adjacencies[i] ) :
				fout.write( ' '+','.join(map(lambda i: str(i), adjacencies[i])))
		except KeyError:
			pass
		fout.write( '\n')

	# -------------------------------------
	# END
	fout.close()

if __name__ == "__main__" :
	run(sys.argv)

