/*
** utils.c
** 
** Made by (Benjamin Negrevergne)
** Email   [firstname].[name]@imag.fr
** 
** Started on  Thu Sep  2 13:15:21 2010 Benjamin Negrevergne
*/

#include <iostream>
#include <fstream>
#include <algorithm>
#include <sstream>
#include <sys/time.h>
#include <iterator>
#include <assert.h>

#include "utils.hpp"
#include "utils_divers.hpp"

using namespace std; 


int set_member_index(const set_t &set, const element_t &e)
{
  int i = 0;
  for(set_t::const_iterator it = set.begin(); it  != set.end(); ++it,++i)
  {
    //    if(element_equal(*it, e))
    if(*it == e)
      return i;
  }
  return -1;
}


bool set_member_sorted(const set_t &sorted_set, const element_t &e)
{
  assert(is_sorted(sorted_set));
  for(set_t::const_iterator it = sorted_set.begin();
      it != sorted_set.end() && *it <= e; ++it){
    if(*it == e)
      return true;
  }
  return false;
}


bool set_member(const set_t &set, const element_t &e)
{
  for(set_t::const_iterator it = set.begin(); it  != set.end(); ++it){
    if(*it == e)
      return true;
  }
  return false;
}



void set_print(const set_t &set)
{
  if(set.size() == 0)
  {
    cout<<"{emptyset}"<<endl;
    return ;
  }

  for(unsigned int i = 0 ; i < set.size(); i++)
  {
    element_print(set[i]);
    cout<<" : ";
  }
  cout<<endl;
}


void print_ot(const Tot &ot)
{
  for(unsigned int i = 0; i < ot.size(); i++)
  {
    cout << "item : " << i << " : " ;

      Toccurences::iterator it ;

      for(it = ot[i].begin() ; it != ot[i].end() ; it++)
      {
        cout << *it << " ";
      }

      cout << endl  ;
  }
}


/**
 * affiche le pattern et un message
 */
void simple_pattern_print(string message, const set_t &set,
						  const vector<string>& items_names)
{
  cout << message << " : " ;

  if(set.size() == 0)
    cout<<"{emptyset}" ;
  else
    for(unsigned int i = 0 ; i < set.size(); i++)
    {
      original_element_print(set[i], items_names);
      cout<< "[" << set[i] << "] " ;
    }

  cout << endl ;

}


void pattern_print(const set_t &set, const set_t &tids)
{

  if(set.size() == 0)
  {
    cout<<"{emptyset}"<<endl;
  }
  else
    for(int i = 0 ; i < set.size(); i++)
    {
      element_print(set[i]);
      cout<<" ";
    }

#ifdef TRACK_TIDS
  if (tids.size() > 0)
  {
    cout << " [ " ;
    for (int i=0 ; i < tids.size() ; ++i)
      cout << tids[i] << " " ;
    cout << "]" ;
  }
#endif //TRACK_TIDS
  cout << endl ;

}


void get_pattern_name(const set_t& pattern,
					  const vector<string>& item_names,
					  string& pattern_name)
{
  pattern_name = "" ;

  int n = pattern.size() ;

  if (n > 0)
  {

		vector<string> attributes_names ;

							  // recupere les noms des attributs

	  for(int i = 0 ; i < n ; i++)
	  {
		  string attribute_name ;
		  get_original_element(pattern[i], item_names, attribute_name);

		  // cout << "attribut : " << attribute_name << endl ;
		  attributes_names.push_back(attribute_name) ;
	  }

							  // insere les noms des attributs dans un string
							  // avec separateur (on profite de l'ordre du set

	  pattern_name = "" ;

	  // cout << "size = : " << pattern.size() << endl ;

	  vector<string>::iterator it = attributes_names.begin() ;


	  for(int i = 0 ; i < n-1 ; i++)
	  {
		  // cout << "*it : " << *it << endl ;
		  pattern_name += *it + "_" ;
		  it++ ;
	  }

	  pattern_name +=  *it ;

   }
}


void original_pattern_print2(const set_t &set,
							 const vector<string>& item_names,
		 	 	 	 	 	 const set_t& permutations)
{
	vector<string> description ;

			// recupere le vecteur<string> correspondant au pattern set

	for(unsigned int i = 0 ; i < set.size() ; i++)
	{
		  description.push_back(item_names[permutations[set[i]]]) ;
	}

			// tri le vecteur
	sort(description.begin(), description.end()) ;

			// affiche le vecteur trié

	// displayVector("description", description) ;


	int n = description.size() ;

	cout << "{" ;

	if (n > 0 )
	{
		for(int i = 0 ; i < n-1 ; i++)
		{
		  cout << "\"" ;
		  cout << description[i] ;
		  cout << "\", " ;
		}

		cout << "\"" ;
		cout << description[n-1] ;
		cout << "\"" ;
	}

    cout << "}" ;
}


void original_pattern_print(const set_t &set, const vector<string>& item_names)
{

  int n = set.size() ;

  cout << "{" ;

  if (n > 0 )
  {
	for(unsigned int i = 0 ; i < set.size()-1 ; i++)
	{
	  cout << "\"" ;
	  original_element_print(set[i], item_names);
	  cout << "\", " ;
	}

	cout << "\"" ;
	original_element_print(set[set.size()-1], item_names);
	cout << "\"" ;
  }

 cout << "}" ;
}


void original_pattern_print_cerr(const set_t &set, const vector<string>& item_names)
{

  int n = set.size() ;

  cerr << "{" ;

  if (n > 0 )
  {
	for(unsigned int i = 0 ; i < set.size()-1 ; i++)
	{
	  cerr << "\"" ;
	  original_element_print_cerr(set[i], item_names);
	  cerr << "\", " ;
	}

	cerr << "\"" ;
	original_element_print_cerr(set[set.size()-1], item_names);
	cerr << "\"" ;
  }

 cerr << "}" ;
}



void pattern_print(const set_t &set, int u_data,
		           const set_t &tids)
{

  if(set.size() == 0)
  {
    cout<<"{ }"<<endl;
  }
  else
    for(int i = 0 ; i < set.size(); i++)
    {
      element_print(set[i]); 
      cout<<" "; 
    }  
  cout<<"("<<u_data<<")" ;

  cout << endl ;

}


void set_print_raw(const set_t &set)
{
  if(set.size() == 0){
    cout<<"{emptyset}"<<endl; 
    return ; 
  }
  for(int i = 0 ; i < set.size(); i++){
    cout<<set[i]; 
    cout<<" : "; 
  }
  cout<<endl; 
}


bool set_equal(const set_t &s1, const set_t &s2)
{
  if(s1.size() != s2.size())
    return false; 
  set_t::const_iterator s1_end = s1.end(); 
  for(set_t::const_iterator it1 = s1.begin(), it2 = s2.begin(); it1 != s1_end; ++it1, ++it2)
  {
    if(*it1 != *it2)
      return false; 
  }
  return true; 
}

bool set_equal_limited(const set_t &s1, int l1, const set_t &s2, int l2)
{
  if(l1 != l2)
    return false; 

  set_t::const_iterator end1 = s1.begin()+l1;
  set_t::const_iterator end2 = s2.begin()+l2; 

  for(set_t::const_iterator it1 = s1.begin(), it2 = s2.begin(); it1 != end1; ++it1, ++it2){
    if(*it1 != *it2)
      return false; 
  }
  return true; 
}



/** \brief returns true if the two sets are equals ignoring element in excluded*/
bool set_equal_with_excluded_elements(const set_t &s1, const set_t &s2,
				      const std::vector<bool> &excluded)
{
    
  set_t::const_iterator end1 = s1.end();
  set_t::const_iterator end2 = s2.end(); 

  set_t::const_iterator it1 = s1.begin(), it2 = s2.begin();
  while(it1 != end1 && it2 != end2){
    if(excluded[*it1]){
      ++it1; continue; 
    }
    if(excluded[*it2]){
      ++it2; continue; 
    }
    if(*it1 != *it2)
      return false; 
    ++it1; ++it2;
  }
  
  for(;it1 != end1;++it1)
    if(!excluded[*it1])
      return false; 
  
  for(;it2 != end2;++it2)
    if(!excluded[*it2])
      return false; 

  return true; 
}

/** \brief returns wether the first set is smaller than the other
    prioritizing elements that are not exlcuded. 
    
    \warning Elements in transactions must be sorted in each category
    (el and non-el). 
*/

bool set_compare_with_excluded_elements(const set_t &s1, const set_t &s2,
					const std::vector<bool> &excluded)
{

  assert(false); 
  /* First compare non-el elements */
  set_t::const_iterator end1 = s1.end();
  set_t::const_iterator end2 = s2.end(); 

  set_t::const_iterator it1 = s1.begin(), it2 = s2.begin();

  
  while(it1 != end1 && it2 != end2){
    if(excluded[*it1]){
      ++it1; continue; 
    }
    if(excluded[*it2]){
      ++it2; continue; 
    }
    if(*it1 < *it2)
      return true; 

    else if(*it1 > *it2)
      return false; 

    ++it1; ++it2;
  }
  
  for(;it1 != end1;++it1)
    if(!excluded[*it1])
      return false; 
  
  for(;it2 != end2;++it2)
    if(!excluded[*it2])
      return true;
  
  /* now compare el-elements. */
  it1 = s1.begin(), it2 = s2.begin();

    while(it1 != end1 && it2 != end2){
    if(!excluded[*it1]){
      ++it1; continue; 
    }
    if(!excluded[*it2]){
      ++it2; continue; 
    }
    if(*it1 < *it2)
      return true; 

    else if(*it1 > *it2)
      return false; 

    ++it1; ++it2;
  }
  
  for(;it1 != s1.end();++it1)
    if(excluded[*it1])
      return false; 
  
  for(;it2 != s2.end();++it2)
    if(excluded[*it2])
      return true;

  /* set are equals */
  return false; 
}


int set_lexical_compare(const set_t &t1, const set_t &t2)
{
  set_t::const_iterator tp = t1.begin(), tpEnd = t1.end();   
  set_t::const_iterator op = t2.begin(), opEnd = t2.end(); 
  int i = 0; 
  for(; tp < tpEnd && op < opEnd; ++tp, ++op, i++){
    if(*tp == *op){
      //      if(unlikely (*tp == prefixBound))
      //	return i+10; /* if the transactions are equals up to the prefix */
      continue; 
    }
    if(*tp > *op)
      return 1; 
    else
      return -1; 
  }
    
  if(tp == tpEnd && op == opEnd)
    return 0; 
  else
    if(tp != tpEnd) 
      return 1; 
    else
      return -1; 
}

int set_lexical_compare_limited(set_t::const_iterator s1, int end1, 
								set_t::const_iterator s2, int end2)
{
  set_t::const_iterator s1end = s1+end1; 
  set_t::const_iterator s2end = s2+end2; 

  int i = 0; 
  for(; s1 < s1end && s2 < s2end; ++s1, ++s2, i++){
    if(*s1 == *s2){
      //      if(unlikely (*s1 == limit_value))
      //      	return i+10; /* if the transactions are equals up to the prefix */
      continue; 
    }
    if(*s1 > *s2)
      return 1; 
    else
      return -1; 
  }
    
  if(s1 == s1end && s2 == s2end)
    return 0; 
  else
    if(s1 != s1end) 
      return 1; 
    else
      return -1; 
}


void set_truncate_above(set_t *s, element_t v)
{
  assert(is_sorted(*s)); 
  int idx = s->size() - 1; 
  while(idx >= 0 && (*s)[idx] > v)
    idx--; 
  s->resize(idx+1);
}


/*
 * return the closed pattern of the pattern set.
 */
set_t support_based_closure(
							const set_t &set,			// pattern à clore
						    int set_support,			// support augmentation
							const SupportTable &support // supports des items n'apparaissant pas
														// dans le pattern a clore
							)
{
  set_t c(set);
  	  	  	  	  	  	  	  // ajoute au pattern à clore
  	  	  	  	  	  	  	  // les items (qui n'appartiennent pas à pattern)
  	  	  	  	  	  	  	  // et qui sont representés dans les transactions
  	  	  	  	  	  	  	  // dont le support est égal à set_support

  for(unsigned int item = 0; item < support.size(); item++)
  {
    if(support[item] == set_support)
      c.push_back(item);
  }
  return c; 
}

void set_insert_sorted(set_t *sorted_set, const element_t e)
{
  /* TODO improve efficiency */
  sorted_set->push_back(e); 
  set_t::iterator xlit = sorted_set->begin(); 
  while(xlit != sorted_set->end()){
    if(*xlit > e){
      std::swap(*xlit, (sorted_set->back())); 
    }
    ++xlit;
  }
 
}



/**
 *	swap indices and values in the vector permutations
 */


void reverse_permutations(set_t *permutations)
{
  vector<bool> done(permutations->size(), false); 

  for(int i = 0; i < done.size(); i++)
  {
    if(done[i] == false)
    {
      int current_idx = i, current_value = (*permutations)[i];

      while(current_value != -1 && done[current_value] == false)
      {
    	  done[current_value] = true;
    	  int tmp = (*permutations)[current_value];
    	  (*permutations)[current_value] = current_idx;
    	  current_idx = current_value;
    	  current_value = tmp;
      }      
    }     
  }  
}



#ifdef TIMING
using std::ofstream; 
static vector<ofstream *> trace_file_fds;
static struct timeval *start_time = NULL;


static void trace_check_file_exists(int tid)
{
      
  if(trace_file_fds.size() > tid && trace_file_fds[tid] != NULL){
    return; 
  }
  else{
    /* unlikely */
    trace_file_fds.resize(tid + 1, NULL); 
    std::ostringstream oss; 
    oss<<"TRACE_"<<tid<<".dat"; 
    trace_file_fds[tid] = new ofstream(oss.str().c_str()); 
  }
}
#endif //TIMING




void trace_timestamp_print(const string &info, int start_end)
{
#ifdef TIMING

  struct timeval tv;
  gettimeofday(&tv, NULL);
  
  int thread_id =  m_thread_id();

  ostream *os = trace_file_fds[thread_id];

  *os<<((tv.tv_sec + (double)tv.tv_usec/1000000.) - (start_time->tv_sec + (double) start_time->tv_usec/1000000.))*1000<<", "; 

  switch(start_end){
  case (EVENT_START):
    *os<<thread_id<<", "<<"START "<<info<<endl; 
    break; 
  case (EVENT_END):
   *os<<thread_id<<", "<<"END "<<info<<endl; 
    break;
  default:
    *os<<thread_id<<", "<<info<<endl; 
    break; 
  }

#endif //TIMING
}



void set_to_bit_representation(const set_t &set, element_t max_element, 
			       std::vector<bool> *set_bit)
{
  set_bit->resize(max_element+1,  false);
  for(set_t::const_iterator it = set.begin(); it != set.end(); ++it)
    if(*it <= max_element)
      (*set_bit)[*it] = true; 
}


bool set_permuted_limited_compare(const set_t &s1, const set_t &s2, 
				  const set_t &permutations, element_t limit)
{

    
  set_t::const_iterator end1 = s1.end();
  set_t::const_iterator end2 = s2.end(); 

  set_t::const_iterator it1 = s1.begin(), it2 = s2.begin();
  while(it1 != end1 && it2 != end2){
    element_t v1 = permutations[*it1]; 
    element_t v2 = permutations[*it2]; 
    if(v1 >= limit)
      {++it1; continue;}

    if(v2 >= limit)
      {++it2; continue;}

    if(v1 > v2)
      return false; 

    if(v1 < v2)
      return true;
    
    ++it1;++it2; 
  }
  
  for(;it1 != end1; ++it1)
    if(permutations[*it1] < limit)
      return false; /* s1 is larger */

  for(;it2 != end2; ++it2)
    if(permutations[*it2] < limit)
      return true; /* s2 is larger */

  return false; /* both are equal */
}


void elsort_transaction(Transaction *t, int max_element, const set_t &el)
{
  /* only for debug purposes, very inefficient */ 
  vector<bool> el_bit(max_element, false);
  for(set_t::const_iterator it = el.begin(); it != el.end(); ++it)
    el_bit[*it] = true; 

  Transaction::iterator last = t->end(); 
  Transaction::iterator it = t->begin(); 
  if(last == it) 
    return;
  last--; 
  while (it <= last){
    if(el_bit[*it]){
      iter_swap(it, last); 
      --last; 
    }
    else
      ++it; 
  }
  t->limit = t->size() - (t->end() - last)+1; 
  std::sort(t->begin(), t->begin()+t->limit); 
  std::sort(t->begin()+t->limit, t->end()); 
}
