/*
 *	MINERLC6
 * read_write_data2.cpp
 *
 *  Created on: 18 mars 2015
 *      Author: Dominique Bouthinon
 *
 */

#include <cstdlib>
#include <algorithm>
#include <vector>
#include <sstream>
#include <fstream>
#include <iostream>
#include <iterator>
#include <set>
#include <string>
#include <tuple>

#include "read_write_data2.hpp"
#include "types_graph.hpp"
#include "utils_divers.hpp"

using namespace std;

/**
 *  get the int values between the current line and the line containing stop
 */
void get_transactions(
				      ifstream& input,
					  int nb_objects,
				      vector<vector<int> >& values
			         )
{
	cerr << "reading transactions..." << endl ;

								// pour chaque objet (ligne)
	for (int i = 0 ; i < nb_objects ; i++)
	{
		 vector<int>    num_items ;
		 vector<string> lines ;
		 string line ;

		 getline(input, line) ;
		 // cout << line << endl ;

		 split(lines, line, " ")   ;

		 if (lines.size() > 1)
		 {
			 get_ints(lines[1], ',', num_items) ;
		 }

		 values.push_back(num_items) ;
	}
}


// creer les predecesseurs à partir de la liste des successeurs

void creer_predecesseurs(Tadjacences_map& adjacences)
{
						// pour chaque noeud n
for(Tadjacences_map_iterator n = adjacences.begin(); n != adjacences.end(); n++)
{
	// cerr << "noeud : " << n->first << endl ;

						// pour chaque successeur s du noeud n
	for (Tvertice_identifiers_set_iterator s = n->second.successors.begin() ;
		 s != n->second.successors.end() ; s++)
	{
		// cerr << "successeur : " << *s << endl ;

						// noeud n (en fait n->first
						// integre les predeccesseurs de s
		adjacences[*s].predeccessors.insert(n->first) ;
	}
}
}


/**
 *  get the adjacences described between the current line and the line containing stop
 *  NOUVEAU TRAITEMENT VALABLE POUR GRAPHE ORIENTE OU NON
 */
void get_adjacences3(
					 ifstream& input,
					 int nb_objects,
					 Tadjacences_map& adjacences
			        )
{
	cerr << "reading adjacences... " << endl ;

	for (int i = 0 ; i < nb_objects ; i++)
	{
		 string line ;

		 getline(input, line) ;
		 // cout << "line = " << line << "xx" << endl ;

		 vector<string> lines ;
		 Tvertice_identifiers_set adjacent_objects ;

		 split(lines, line, " ") ;

		 Tvertice_identifier ident = stoi(lines[0]) ;	// ident objet concerne
		 // cout << ident << endl ;


		 if (lines.size() > 1)
		 {
			 get_ints(lines[1], ',', adjacent_objects) ;
		 }

		 // displaySet("adja", adjacent_objects) ;

		 	 	 	 	 	 	 	 	 	 // recherche le numero de transaction
		 	 	 	 	 	 	 	 	 	 // correspondant à l'objet de nom v[0]
		 Tadjacences adja ;
		 									 // distribue les identifiants des noeuds
		 									 // comme successeurs

		 Tvertice_identifiers_set_iterator it     = adjacent_objects.begin() ;
		 Tvertice_identifiers_set_iterator it_end = adjacent_objects.end() ;

		 // cout << "ident = " << ident << endl ;

		 for ( ; it != it_end ; it++)
		 {
				// cout << "*it = " << *it << endl ;
				adja.successors.insert(*it) ;
		 }

		 adjacences.insert(make_pair(ident, adja)) ;
	}
}



/**
 *  get the adjacences described between the current line and the line containing stop
 */
void get_adjacences2(
					 ifstream& input,
					 int nb_objects,
					 Tadjacences_map& adjacences
			        )
{
	cerr << "reading adjacences..." << endl ;

	for (int i = 0 ; i < nb_objects ; i++)
	{
		 string line ;

		 getline(input, line) ;
		 // cout << "line = " << line << "xx" << endl ;

		 vector<string> lines ;
		 Tvertice_identifiers_set adjacent_objects ;

		 split(lines, line, " ") ;

		 Tvertice_identifier ident = stoi(lines[0]) ;	// ident objet concerne
		 // cout << ident << endl ;


		 if (lines.size() > 1)
		 {
			 get_ints(lines[1], ',', adjacent_objects) ;
		 }

		 // displaySet("adja", adjacent_objects) ;

		 	 	 	 	 	 	 	 	 	 // recherche le numero de transaction
		 	 	 	 	 	 	 	 	 	 // correspondant à l'objet de nom v[0]
		 Tadjacences adja ;
		 									 // distribue les identifiants des noeuds
		 									 // adjacents en predecesseurs et successeurs

		 Tvertice_identifiers_set_iterator it     = adjacent_objects.begin() ;
		 Tvertice_identifiers_set_iterator it_end = adjacent_objects.end() ;

		 // cout << "ident = " << ident << endl ;

		 for ( ; it != it_end ; it++)
		 {
				// cout << "*it = " << *it << endl ;

			if(*it < ident)
				adja.predeccessors.insert(*it) ;
			else
				adja.successors.insert(*it) ;
		 }

		 adjacences.insert(make_pair(ident, adja)) ;
	}
}


void read_data_new_ri(
						bool graphe_oriente,
						const string fileName,
						vector<string>& object_names,
						vector<string>& item_names,
						vector<vector<int> >& transacts,
						Tadjacences_map& adjacences
				 	 )
{
	    ifstream input(fileName , ifstream::in); 		// ouvre fichier

	    if (input)
	    {
			string line;

			getline(input, line) ;						// commentaire
			// cout << line << endl ;

			getline(input, line) ;						// lecture noms objets
			// cout << "line = " << line << endl ;

			split(object_names, line, " |") ;
			trimStringVector(object_names) ;
														// le dernier objet peut être vide
														// si la ligne contenant le nom des objets
														// finissait par |
			int nb_objects = object_names.size() ;
			string& last_object = object_names[nb_objects-1] ;

			if (last_object.size() == 0)
			{
				object_names.pop_back() ;
			}

			nb_objects = object_names.size() ;

			// displayVector("object_names", object_names);
			// cout << endl ;

			getline(input, line) ;						// lecture  attributs
			// cout << line << endl ;

			split(item_names, line, " |") ;
			trimStringVector(item_names)  ;

												// le dernier item peut être vide
												// si la ligne contenant les items (attributs)
												// finissait par "|"
			int nb_attributes = item_names.size() ;
			string& last_attribute = item_names[nb_attributes-1] ;

			if (last_attribute.size() == 0)
			{
				item_names.pop_back() ;
			}

			// displayVector("item_names", item_names);
			// cout << endl ;
												// lecture des transactions
			get_transactions(input, nb_objects, transacts) ;
			// displayVectorVector("transacts", transacts) ;

												// lecture #
			getline(input, line) ;
			// cout << line << endl ;

												// lecture des adjacences
			// get_adjacences2(input, nb_objects, adjacences) ;
			// afficher_adjacences(adjacences) ;
			if (graphe_oriente)
			{
				get_adjacences3(input, nb_objects, adjacences) ;
				creer_predecesseurs(adjacences) ;
			}
			else
				get_adjacences2(input, nb_objects, adjacences) ;

			input.close() ;

	    }
	    else
	    {
	    	cout << "the file " << fileName << " cannot be opened." << endl ;
	    }
}



// lit graphe au nouveau format
void read_graph_data2(
		  	  bool graphe_oriente,
			  const string fileName,
			  vector<string>& object_names,
			  Tadjacences_map& adjacences
		      )
{
ifstream input(fileName , ifstream::in); 		// ouvre fichier

if (input)
{
	string line;

	getline(input, line) ;						// lecture commentaire
	// cout << "line = " << line << endl ;

	getline(input, line) ;						// lecture noms objets
	// cout << "line = " << line << endl ;

	split(object_names, line, " |") ;
	trimStringVector(object_names) ;
												// le dernier objet peut être vide
												// si la ligne contenant le nom des objets
												// finissait par |
	int nb_objects = object_names.size() ;
	string& last_object = object_names[nb_objects-1] ;

	if (last_object.size() == 0)
	{
		object_names.pop_back() ;
	}

	nb_objects = object_names.size() ;

	getline(input, line) ;						// lecture  attributs
	// cout << line << endl ;
	
											// lecture des transactions

								// pour chaque objet (ligne)
	for (int i = 0 ; i < nb_objects ; i++)
	{
		 
		 string line ;

		 getline(input, line) ;
		 // cout << line << endl ;
	}

										// lecture #
	getline(input, line) ;
	// cout << line << endl ;
											
	if (graphe_oriente)
	{
		get_adjacences3(input, nb_objects, adjacences) ;
		creer_predecesseurs(adjacences) ;
	}
	else
		get_adjacences2(input, nb_objects, adjacences) ;


	// afficher_adjacences(adjacences) ;

}
else
{
	cout << "the file " << fileName << " cannot be opened." << endl ;
}

// verifier après getAdjacences la différence entre objets traités
// et objets non traités (en utilisant les indices)
// ajouter successeurs vide et predecesseurs vides aux objets non traités
}






