/**
 *	SUPPORTS_ATTENDUS.CPP 
 *	program modified by Dominique Bouthinon (LIPN, University Paris 13)
 *
 * THIS PROGRAM MUST BE COMPILED WITH VERSION C++11
 * WITH OPTIONS THAT ALLOW FOR <functionial>
 *
 */  
#include <cstdlib>
#include <algorithm>
#include <vector>
#include <sstream>
#include <fstream>
#include <iostream>
#include <sys/time.h>
#include <time.h>
#include <ctime>
#include <stdio.h>
#include <unistd.h>
#include <set>
#include <string>
#include <tuple>
#include <math.h>

#include "statsModularity.hpp"
#include "read_supports_input.hpp"
#include "graph.hpp"
#include "directed_graph.hpp"
#include "types_graph.hpp"
#include "read_write_data2.hpp"
#include "parameters.hpp"
#include "common_functions.hpp"
#include "common_supports_functions.hpp"

using std::cout;
using std::cerr;
using std::endl;

using namespace std;

			//	gestion des options du programme
bool handle_supports_attendus_parameters(
					                     int argc,
					                     char* argv[],
										 bool& modularity_on,
										 float modularity,
					                     abstraction& a,
								         bool& directed_graph,
					                     int& n
					                )
{
	bool ok = true ;

	int i = 3 ;							// 3 premiers parametres vus avant
	while (i < argc && ok)
	{
        if (handle_abstractions(argv, a, i, "1")) {}
		else if(string(argv[i]) == "-n")
		{
			n = atoi(argv[i+1]) ;
			i = i + 2 ;
		}
		else if(string(argv[i]) == "-o")
		{
			directed_graph = true ;
			i++ ;
		}
		else if(string(argv[i]) == "-lm")
		{
			modularity = atof(argv[i+1]) ;
			modularity_on = true ;
			i = i + 2 ;
		}
		else
		{
			ok = false ;
			cerr << "parameter " << argv[i] << " not recognized" << endl ;
		}
	}

	return ok ;
}


			// !!! pas utilise !!!
void write_final_messages(statsModularity& statsMod,
					     int nb_closed_patterns, int nb_rejected_closed_patterns,
						 int num_pattern, time_t& debut, time_t & fin)
{
		
	 cout << "# ===================================================================" << endl ;
	 cout << nb_closed_patterns          << "# closed pattern(s)"              	   << endl ;
	 cout << nb_rejected_closed_patterns << "# rejected closed pattern(s)" 	       << endl ;
	 cout << num_pattern                 << "# triplet(s) (c, e, l)."    		   << endl ;
	 cout << "# nec        = " << statsMod.getNec() << endl ;
	 cout << "# f          = " << statsMod.getF()   << endl ;
	 cout << "# lme        = " << statsMod.getNbLme() << endl ;
	 cout << "# lm         = " << statsMod.getNbLm() << endl ;
	 cout << "# begin      = " << ctime(&debut) ;
	 cout << "# end        = " << ctime(&fin)  ;
	 cout << "# duration   = " << difftime(fin,debut) << " seconds." << endl ;
}


void sauvegarder_supports(
					  	  vector<tuple<string,int,int> >& supports_observes,
						  double supports_attendus[],
						  double ecarts_types[]
				     	 )
{
	int nb_supports = supports_observes.size() ;

	for (int i = 0 ; i < nb_supports ; i++)
	{
		cout  << get<0>(supports_observes[i])<< ", "
			  << get<1>(supports_observes[i])<< ", "
			  << get<2>(supports_observes[i])<< ", "
			  << supports_attendus[i] << ", "
			  << ecarts_types[i] << endl ;
	}
}

				// compare deux tuples selon son deuxième argument
struct sort_pred
{
    bool operator()(const tuple<string, int,int> &left, const tuple<string, int,int> &right)
    {
        return get<1>(left) > get<1>(right);
    }
} ;


// ------------------------------- parameters ----------------------------------


void write_supports_attendus_parameters(string program,
					               string graph_data_file,
					               string supports_data_file,
								   bool modularity_on, 
				  				   float modularity,
					               abstraction& a1,
					               int n)
{
  cout << "# program                          = " << program		  	  << endl ;
  cout << "# graph data file                  = " << graph_data_file 	  << endl ;
  cout << "# supports data file               = " << supports_data_file   << endl ;
  cout << "# nombre de tirages                = " << n 					  << endl ;
  cout << "# local modularity                 = " << modularity_on		  <<endl ;
  cout << "# local modularity  threshold      = " << modularity			  << endl ;
  write_abstractions(a1, "(a1)") ;
  cout << "# " << endl ;
}


// -------------------------------- CALCUL DES SUPPORTS ATTENDUS ---------------

void calcul_supports_attendus(
					   	   	   graph<string>* initial_graph,	// graphe initial
							   bool graphe_dirige,
							   vector<tuple<string,int,int> >& supports_observes,	// (par ordre croissant)
							   int n,							// nombre d'iterations
							   abstraction a,
							   double* supports_attendus,
							   double* ecarts_types
			    	 	 	 )
{
													// chaque triplet de supports_observes contient :
													// element 0 : clos abstrait
													// element 1 : support non abstrait observé du clos
													// element 2 : support abstrait observé du clos


	srand(time(NULL)) ;								// initialisation séquence aleatoire

	int nb_supports = supports_observes.size() ;
													// met les identificateurs des noeuds de g
													// dans un tableau
	int nb_vertices = initial_graph -> nb_vertices() ;
	// cerr << "nb noeuds = " << nb_vertices << endl ;
	Tvertice_identifier vertices[nb_vertices] ;

	int id = 0 ;

	Tvertice_identifiers_set_iterator  it     = initial_graph->get_vertice_identifiers_list().begin()  ;
	Tvertice_identifiers_set_iterator  it_end = initial_graph->get_vertice_identifiers_list().end()  ;

	for ( ; it != it_end ; it++)
	{
		vertices[id] = *it ;
		// cout << vertices_g[id] << endl ;
		id++ ;
	}

	double* sommes_supports_attendus        = new double[nb_supports] ;
	double* sommes_carres_supports_attendus = new double[nb_supports];

													// initialisation sommets
	for (int k = 0 ; k < nb_supports ; k++)
	{
		  sommes_supports_attendus[k] = 0 ;
		  sommes_carres_supports_attendus[k] =  0 ;
	}

													// faire n tirages
	for (int i = 0 ; i < n ; i++)
	{
		// cerr << "tirage  " << i << endl ;
		
		graph<string>* g1 ;

   								// copie du graphe initial
		if (graphe_dirige)
		{
			directed_graph<string>* gp = dynamic_cast<directed_graph<string>* > (initial_graph) ;
			g1 = new directed_graph<string>(*gp) ;
		}
		else
			g1 = new graph<string>(*initial_graph) ;

		// g1-> display_vertice_identifiers() ;
		int limite_sup_tirage = nb_vertices  ;
				 									// pour chaque support observe (du plus grand au plus petit)
													// (j identifie un clos abstrait, et un support)
		for (int j = 0 ; j < nb_supports ; j++)
		{
													// on tire les noeuds à exclure
			int nb_noeuds_a_tirer = limite_sup_tirage - get<1>(supports_observes.at(j)) ;

			Tvertice_identifiers_set* drawn_vertices = new Tvertice_identifiers_set ;

			// Tvertice_identifiers_set drawn_vertices ;

													// tire au hasard supports[j] noeuds
													// et les place dans drawn_vertices
													// ce sont les noeuds à exclure
			// cerr << "j = " << j << endl ;
			// cout << "motif = " << get<0>(supports_observes.at(j)) << endl ;
			// cout << "support non abstrait observe = " << get<1>(supports_observes.at(j)) << endl ;
			// cout << "support abstrait observe = " << get<2>(supports_observes[j]) << endl ;

			// cout << "nb_vertices_g = " << nb_vertices << endl ;
			// cout << "limite_sup_tirage= " << limite_sup_tirage << endl ;
			// cout << "nb_noeuds_a_tirer = " << nb_noeuds_a_tirer << endl ;

			//displayTab(vertices, nb_vertices) ;
			// displayTab(vertices, limite_sup_tirage) ;

			get_random_vertices(nb_noeuds_a_tirer, vertices, limite_sup_tirage, *drawn_vertices) ;

			// cout << "limite_sup_tirage= " << limite_sup_tirage << endl ;

			// cout << "apres tirage : " << endl ;
			// displayTab(vertices, nb_vertices) ;
			// displayTab(vertices, limite_sup_tirage) ;

			// graph<string>::display_vertice_identifiers_set(drawn_vertices) ;
			// cout << endl ;

			// cout << "noeuds g1 avant effacement " << endl ;
			// g1->display_vertice_identifiers() ;
													// on élimine les noeuds à exclure
			g1->discard_vertices(*drawn_vertices) ;

			delete (drawn_vertices) ;

			// cout << "noeuds g1 apres effacement " << endl ;
			// g1->display_vertice_identifiers() ;

			graph<string>* gA ;
												// copie du graphe initial
			if (graphe_dirige)
			{
				directed_graph<string>* gp =  dynamic_cast<directed_graph<string>* > (g1) ;
				gA = new directed_graph<string>(*gp) ;
			}
			else
				gA = new graph<string>(*g1) ;


			// gA->display_graph() ;
			// cout << endl << "ga (avant abstraction) nombre noeuds = " << gA->nb_vertices() << endl ;

			Troles_set roles ;
			gA->apply_abstractions(a, 1, initial_graph, roles) ;	// abstraction avec min_sup = 1

			// gA->display_graph() ;
			// cout << endl << "ga nombre noeuds = " << gA->nb_vertices() << endl ;

													// suport abstrait attendu
			sommes_supports_attendus[j]        += gA->nb_vertices() ;
			sommes_carres_supports_attendus[j] += gA->nb_vertices() *
												  gA->nb_vertices() ;

			// cout << "sommes_supports = " << sommes_supports_attendus[j] << endl ;
			// cout << "sommes_carress = " << sommes_carres_supports_attendus[j] << endl ;

			delete (gA) ;
		}
		// fin j (supports)

		delete (g1) ;
	}
	// fin i (tirages)
												// calcul des supports abstraits attendus
												// a partir des sommes
	for (int j = 0 ; j < nb_supports ; j++)
	{
		// cout << endl << "motif = " << get<0>(supports_observes.at(j)) << endl ;
		// cout << "sommes_supports = " << sommes_supports_attendus[j] << endl ;
		// cout << "sommes_carres = " << sommes_carres_supports_attendus[j] << endl ;

		supports_attendus[j] = sommes_supports_attendus[j] / (n * 1.0) ;

		// cout << "supports attendus " << supports_attendus[j] << endl ;

        // cout << "carres/n " << sommes_carres_supports_attendus[j] / (n*1.0) << endl ;

        // cout << "carre supports " << supports_attendus[j] * supports_attendus[j] << endl ;

		ecarts_types[j] = sqrt(
								(sommes_carres_supports_attendus[j] / (n *1.0)) -
								(supports_attendus[j] * supports_attendus[j])
							  ) ;

		// cout << "support abstrait observe = " << get<2>(supports_observes.at(j)) << endl ;
 		// cout << "support abstrait attendu = " << supports_attendus[j] << endl ;
		// cout << "ecart type = "      << ecarts_types[j] << endl ;
	}

	delete[] sommes_carres_supports_attendus ;
	delete[] sommes_supports_attendus ;
}


// ============================== MAIN ========================================

int main(int argc, char* argv[])
{
  vector<string> objects_names ;
  vector<vector<int> > initial_objects ;
  Tadjacences_map  adjacences ;
  vector<string> description_noeuds ;					// pas utilisé mais indispensable
  vector<tuple<string, int,int> > supports_observes ;
  float modularity				= 0 ;
  bool modularity_on			= false ;
  statsModularity statsMod ;

  bool local	  				= false ;		// local ou global
  bool communauty_abstraction 	= false ;		// abstraction dans les communautés ounon
  int min_sup	  				= 1     ;		// frequence minimal des patterns a considérer
												// pas de graphe des triangles
  abstraction a1 ;
  abstraction a2 ;
  string name_data_file       	= argv[1]   ;
  bool graphe_dirige			= false ;
  
  string graph_data_file    = argv[1]       ;
  string supports_data_file = argv[2]       ;
  int  n  = 1                               ;   // nombre de repetitions

  if (!handle_supports_attendus_parameters(argc, argv, modularity_on, modularity, a1, graphe_dirige, n))
	  return 0 ;

  write_supports_attendus_parameters(argv[0], graph_data_file, supports_data_file,
		                             modularity_on, modularity, a1, n) ;

  cerr << endl << "reading graph data..." << endl ;


  read_graph_data2(graphe_dirige, graph_data_file, objects_names, adjacences) ;
  // afficher_adjacences(adjacences) ;


 cerr << endl << "reading supports..." << endl ;

 read_supports_data(supports_data_file, supports_observes) ;

  // displayVectorTuples("supports", supports_observes) ;
  sort(supports_observes.begin(), supports_observes.end(), sort_pred()) ;
  // displayVectorTuples("supports", supports_observes) ;

 directed_graph<string > dg ;

 // cout << "creer graphe initial" << endl ;

  graph<string>* g = create_initial_graph(graph_data_file, false, objects_names,
				 	 	 	 	 	 	  description_noeuds, adjacences, graphe_dirige, a1, a2) ;

  // g->display_graph() ;

  // cout << "fin creer graphe initial " << endl ;

  // displayVector("objects_names", objects_names) ;
  // cout << endl ;

  int nb_supports = supports_observes.size() ;

  // cout << "nb_supports = " << nb_supports << endl ;

  double* supports_attendus = new double [nb_supports] ;
  double* ecarts_types      = new double [nb_supports] ;

  time_t debut ;
  time(&debut) ;

  try
  {
	  cerr << "computing supports..." << endl ;

	  calcul_supports_attendus(g, graphe_dirige, supports_observes, n, a1,
			  	  	  	       supports_attendus, ecarts_types) ;
  	  time_t fin ;
  	  time(&fin) ;

  	  cout << "#" << endl ;
  	  cout << "# begin      = " << ctime(&debut) ;
  	  cout << "# end        = " << ctime(&fin)  ;
  	  cout << "# duration   = " << difftime(fin,debut) << " seconds." << endl ;
  }
  catch(exception& e)
  {
	  cerr << e.what() << endl ;
  }
  catch(string& s)
  {
	  cerr << s << endl ;
  }

  delete (g) ;

  sauvegarder_supports(supports_observes, supports_attendus, ecarts_types) ;

  delete[] ecarts_types ;
  delete[] supports_attendus ;


  cerr << "end of the program " << argv[0] << endl ;

  return 0 ;
}
