
#include "../include/read_supports_input_bi.hpp"

#include <cstdlib>
#include <algorithm>
#include <vector>
#include <sstream>
#include <fstream>
#include <iostream>
#include <iterator>
#include <set>
#include <string>
#include <tuple>
#include <map>

#include "types_graph.hpp"
// #include "utils_divers.hpp"

using namespace std;


void extrait_extension_support(string& line,
							   set<int>& VO,
							   set<int>& VI,
							   map<string, Tvertice_identifier>& vertice_idents,
							   int debut,
							   string& extension,
							   int& fin,
							   int& support_out,
							   int& support_in)
{

	// displaySet("VO", VO) ;
	// displaySet("VI", VI) ;

	extrait_motif(line, debut, extension, fin) ;
	// cout << "extension A = " << extension << endl ;

	string extension2 = extension.substr(1, extension.size()-2) ;
	// cout << "extension2 = " << extension2 << endl ;

								// elimine les espaces
	extension2.erase(remove(extension2.begin(), extension2.end(), ' '), extension2.end());
								// elimine les guillemets
	extension2.erase(remove(extension2.begin(), extension2.end(), '"'), extension2.end());

	// cout << "extension2 =" << extension2 << endl ;
							   //
	vector<string> items ;
	split(items, extension2, ",") ;
	sort(items.begin(), items.end()) ;

	// displayVector("items", items) ;

	set<Tvertice_identifier> idents ;

							// retrouve les idenficateurs des noeuds
							// a partir de leur libelle

		 // pour chaque noeud x
	for (int i  = 0  ; i < items.size() ; i++)
	{
		// cout << "item :" << i << " =xxx" << items[i] << "xxx"<< endl ;
		// cout << "yyy " << vertice_idents[items[i]] << endl ;

		Tvertice_identifier ident = vertice_idents[items[i]] ;

		// cout << "ident = " << ident << endl ;
		idents.insert(ident) ;
	}

	// displaySet("idents", idents) ;


	set<Tvertice_identifier> out ;
	set<Tvertice_identifier> in  ;

	set_intersection (
				       idents.begin(), idents.end(),
					   VO.begin(), VO.end(),
				       inserter(out, out.begin())
				      ) ;

	set_intersection (
					   idents.begin(), idents.end(),
					   VI.begin(), VI.end(),
					   inserter(in, in.begin())
					      ) ;

	support_out = out.size() ;
	support_in  = in.size()  ;

	// displaySet("out", out) ;
	// displaySet("in", in) ;
}


	// idem extrait_extension_support mais en tenant compte de la presence
	// de roles ({role1->{...}, role2->{...},...}
void extrait_extension_support_roles(string& line, int debut,
							   	    int& support_abstrait)
{
				// recherche la position de l'accolade terminale
	int fin_accolade =  line.size()-1 ;

	int debut_role = debut + 1 ;
	int fin_role ;
	string role  ;

	support_abstrait = 0 ;

				// recherche l'extension liee a chaque role
	do
	{
		int accolade_ouvrante = line.find('{', debut_role) ;
		extrait_motif(line, accolade_ouvrante, role, fin_role) ;

		// cout << "role = " << role << endl ;

		string role2 = role.substr(1, role.size()-2) ;

		// cout << "role2 = " << role2 << endl ;

		vector<string> items_role ;

		split(items_role, role2, ",") ;

		// displayVector("items_roles", items_role) ;

		support_abstrait = support_abstrait + items_role.size() ;

		debut_role = fin_role + 2 ;
	}
	while(debut_role < fin_accolade) ;
}

				// recherche le motif et calcul les supports à partir des
				// extensions
void traite_ligne_supports(string& line,
						   set<Tvertice_identifier>& VO,
						   set<Tvertice_identifier>& VI,
						   map<string, Tvertice_identifier>& vertice_idents,
						   string& motif,
		                   string& extension,
						   int& support_out,
						   int& support_in,
						   string& extension_abstraite,
						   int& support_abstrait)
{

	// cout << line << endl ;
												// isole le motif
	int debut_motif = 0 ;
	int fin_motif ;
	extrait_motif(line, debut_motif, motif, fin_motif) ;

	// cout << "motif = " << motif << endl ;

												// isole l'extension non abstraite
												// sans les accolades
	int debut_extension  = fin_motif + 3 ;
	int fin_extension ;


	extrait_extension_support(line, VO, VI, vertice_idents, debut_extension, extension,
						      fin_extension, support_out, support_in) ;

	// cout << "extension = " << extension << endl ;

												// isole l'extension non abstraite
												// sans les accolades
	int debut_extension_abstraite  = fin_extension + 3 ;
	int fin_extension_abstraite ;
	extrait_extension_support_roles(line, debut_extension_abstraite,
							        support_abstrait) ;
}


void read_supports_data(const string fileName,
						set<Tvertice_identifier>& VO,
						set<Tvertice_identifier>& VI,
						map<string, Tvertice_identifier>& vertice_idents,
						vector<tuple<string, int, int, int> >& supports_observes)
{
		ifstream input(fileName , ifstream::in); 		// ouvre fichier

	    if (input)
	    {
			string line ;

			while(getline(input, line))
			{
				// cerr << "line = " << line  << endl ;

				if (line.at(0) != '#')
				{
					string motif ;
					string extension ;
					string extension_abstraite ;
					int support_out  ;
					int support_in   ;
					int support_abstrait  ;

					traite_ligne_supports(line, VO, VI, vertice_idents, motif, extension,
										  support_out, support_in,
										  extension_abstraite, support_abstrait) ;

					// cout << "motif   = " << motif << endl ;
					// cout << "extension = " << extension << endl ;
					// cout << "support_out = " << support_out << endl ;
					// cout << "support_in = " << support_in << endl ;
					// cout << "extension_abstraite = " << extension_abstraite << endl ;
					// cout << "support_abstrait = " << support_abstrait << endl << endl ;

					supports_observes.push_back(make_tuple(motif, support_out, support_in, support_abstrait)) ;
				}
			}
	    }
	    else
	    {
	    	cout << "the file " << fileName << " cannot be opened." << endl ;
	    }
}
